#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import division
from builtins import range
import warnings
import copy
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import odeint
from scipy.optimize import curve_fit
import pandas as pd
import lmfit
import curveball
import seaborn as sns
sns.set_style("ticks")


def _alfa(t, q0, v):
    if np.isinf(q0) or np.isinf(v):
        return 1.0
    return q0 / (q0 + np.exp(-v * t))

def double_baranyi_roberts_ode0(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model [1]_. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i \Big(1 - \Big(\frac{\sum_{j}{N_j}}{K_i}\Big)^{\nu_i}\Big)

    - :math:`N_i`: population size of strain *i*.
    - :math:`r_i`: initial per capita growth rate of strain *i*
    - :math:`K_i`: maximum population size of strain *i*
    - :math:`\nu_i`: curvature of the logsitic term of strain *i*
    - :math:`\alpha_i(t)= \frac{q_{0,i}}{q_{0,i} + e^{-v_i t}}`
    - :math:`q_{0,i}`: initial adjustment of strain *i* to current environment
    - :math:`v_i`: adjustment rate of strain *i*

    Parameters
    ----------
    y : float, float
        population size
    t : float
        time, usually in hours
    r : float, float
        initial per capita growth rate
    K : float, float
        maximum population size (:math:`K>0`)
    nu : float, float
        curvature of the logsitic term (:math:`\nu>0`)
    q0 : float, float
        initial adjustment to current environment (:math:`0<q_0<1`)
    v : float, float
        adjustment rate (:math:`v>0`)

    Returns
    -------
    float, float
        population growth rate.

    References
    ----------
    .. [1] Baranyi, J., Roberts, T. A., 1994. `A dynamic approach to predicting bacterial growth in food <www.ncbi.nlm.nih.gov/pubmed/7873331>`_. Int. J. Food Microbiol.
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    dydt = alfa[0] * r[0] * y[0] * (1 - ((y[0] + y[1]) / K[0])**nu[0]), alfa[1] * r[1] * y[1] * (1 - ((y[0] + y[1]) / K[1])**nu[1])
    return dydt


def double_baranyi_roberts_ode1(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i \Big(1 - \Big(\sum_{j}{\frac{N_j}{K_j}}\Big)^{\nu_i}\Big)

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    dydt = alfa[0] * r[0] * y[0] * (1 - (y[0] / K[0] + y[1] / K[1])**nu[0]), alfa[1] * r[1] * y[1] * (1 - (y[0] / K[0] + y[1]/K[1])**nu[1])
    return dydt


def double_baranyi_roberts_ode2(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i \Big(1 - \sum_{j}{\Big(\frac{N_j}{K_j}\Big)^{\nu_j}}\Big)

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    dydt = alfa[0] * r[0] * y[0] * (1 - (y[0] / K[0])**nu[0] - (y[1] / K[1])**nu[1]), alfa[1] * r[1] * y[1] * (1 - (y[0] / K[0])**nu[0] - (y[1] / K[1])**nu[1])
    return dydt


def double_baranyi_roberts_ode3(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i \Big( 1 - \Big(\frac{\sum_j{N_j}}{\bar{K}}\Big)^{\nu_i} \Big)

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    Kmean = (K[0] * y[0] + K[1] * y[1]) / (y[0] + y[1])
    dydt = alfa[0] * r[0] * y[0] * (1 - ((y[0] + y[1]) / Kmean)**nu[0]), alfa[1] * r[1] * y[1] * (1 - ((y[0] + y[1]) / Kmean)**nu[1])
    return dydt

def double_baranyi_roberts_ode4(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    dydt = alfa[0] * r[0] * y[0], alfa[1] * r[1] * y[1]
    return dydt


def double_baranyi_roberts_ode5(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i \Big( 1 - \Big(\frac{N_i}{K_i}\Big)^{\nu_i}\Big)

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    dydt = alfa[0] * r[0] * y[0] * (1 - (y[0]/K[0])**nu[0]), alfa[1] * r[1] * y[1] * (1 - (y[1]/K[1])**nu[1])
    return dydt


def double_baranyi_roberts_ode6(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i \Big(1 - \Big(\frac{\sum_j{N_j}}{K_i}\Big)^{\bar{\nu}}\Big)

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    numean = (nu[0]*y[0] + nu[1]*y[1]) / (y[0] + y[1])
    dydt = alfa[0] * r[0] * y[0] * (1 - ((y[0]+y[1])/K[0])**numean), alfa[1] * r[1] * y[1] * (1 - ((y[0]+y[1])/K[1])**numean)
    return dydt


def double_baranyi_roberts_ode7(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_1}{dt} = 0 \;\;\; \frac{dN_2}{dt} = r_2 \alpha_2(t) N_2 \Big(1 - \Big(\frac{N_2}{K_2}\Big)^{\nu_1}\Big)

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    numean = (nu[0]*y[0] + nu[1]*y[1]) / (y[0] + y[1])
    dydt = 0, alfa[1] * r[1] * y[1] * (1 - (y[1]/K[1])**nu[1])
    return dydt


def double_baranyi_roberts_ode8(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model. The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i N_i \Big(1 - \Big(\frac{\sum_{j}{N_j}}{K_i}\Big)^{\nu_i}\Big)

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    dydt = r[0] * y[0] * (1 - ((y[0] + y[1]) / K[0])**nu[0]), r[1] * y[1] * (1 - ((y[0] + y[1]) / K[1])**nu[1])
    return dydt


def double_baranyi_roberts_gimenez_delgado_ode(y, t, r, K, nu, q0, v):
    r"""A two species Baranyi-Roberts model with competition model inspired by Gimenez and Delgado (2004).
    The function calculates the population growth rate at given time points.

    .. math::

        \frac{dN_i}{dt} = r_i \alpha_i(t) N_i \prod_j{\Big(1 - \Big(\frac{N_j}{K_j}\Big)^{\nu_j}\Big)}

    See also
    --------
    curveball.competitions.double_baranyi_roberts_ode0
    """
    alfa = _alfa(t, q0[0], v[0]), _alfa(t, q0[1], v[1])
    prod = (1 - (y[0]/K[0])**(nu[0])) * (1 - (y[1]/K[1])**(nu[1]))
    dydt = alfa[0] * r[0] * y[0] * prod, alfa[1] * r[1] * y[1] * prod
    return dydt


def compete(m1, m2, p0=(0.5, 0.5), y0=None, t=None, hours=24, num_of_points=100, 
            nsamples=1, lag_phase=True, ode=double_baranyi_roberts_ode1,
            params1=None, params2=None, ci=95, colors=None, ax=None, PLOT=False,
            sampler='covar', df1=None, df2=None):
    """Simulate competitions between two strains using growth parameters estimated
    by fitting growth models to growth curves data.

    Integrates a 2-dimensional ODE system given by `ode`
    with parameters extracted from :py:class:`lmfit.model.ModelResult` instances `m1` and `m2`.
    This implementation includes plotting (if required by `PLOT`);
    resampling from the distribution of model parameters (when `nsamples` > 1);
    changing the ODE system (by providing a different function in `ode`);
    and more.

    The function competes two strains/species;
    therefore it expects two :py:class:`lmfit.model.ModelResult` objects,
    two initial values in `y0`, etc.

    Parameters
    ----------
    m1, m2 : lmfit.model.ModelResult
        model fitting results of growth models defined in :py:mod:`curveball.models`.
    p0, y0 : tuple, optional
        `p0` is the initial relative frequencies; `y0` is the initial population sizes.
        if `y0` is given than ``y0[0]`` and ``y0[0]`` are the initial population size for `m1` and `m2`, respectively.
        if `y0` is not given, then it will be set to the average estimated `y0` in `m1` and `m2` multiplied by `p0`.
    t : numpy.ndarray or None, optional
        array of time points in which to calculate the population sizes.
    hours : int, optional
        if `t` is not given, determines how many hours should the competition proceed, defaults to 24.
    num_of_points : int, optional
        if `t` is not given, determines the number of time points to use, defaults to 100.
    nsamples : int, optional
        how many replicates of the competition should be simulated;
        if `nsamples` = 1, only one competition is simulated with the estimated parameters;
        otherwise `nsamples` competitions are simulated with parameters drawn from a parameter distribution
        determined by `sampler`. Defaults to 1.
    lag_phase : bool, optional
        if `True`, use lag phase as given by `m1` and `m2`. Otherwise, override the lag phase parameters to prevent a lag phase. Defaults to :const:`True`.
    ode : func, optional
        an ordinary differential systems system defined by a function that accepts ``y``, ``t``, and additional arguments, and returns the derivate of ``y`` with respect to ``t``. Defaults to :py:func:`.double_baranyi_roberts_ode0`.
    params1, params2 : dict, optional
        dictionaries of model parameter values; if given, overrides values from `m1` and `m2`.
    ci : float, optional
        confidence interval size, in (0, 100), only applicable when `PLOT` is :const:`True`, defaults to 95%.
    colors : sequence of str, optional
        if `PLOT` is :const:`True`, this sets the colors of the drawn lines. `colors[0]` will be used for `m1`; `colors[1]` for `m2`. If not provided, defaults to the current pallete.
    ax : matplotlib.axes.Axes, optional
        if `PLOT` is :const:`True`, an axes to plot into; if not provided, a new one is created.
    PLOT : bool, optional
        if :const:`True`, the function will plot the curves of *y* as a function of *t*. Defaults to :const:`False`.
    sampler : str, optional
        if ``covar``, the parameters will be sampled using the covariance matrix (:py:func:`curveball.models.sample_params`);
        if ``bootstrap`` the parameters will be sampled by resampling the growth curves (:py:func:`curveball.models.bootstrap_params`).
    df1, df2 : pandas.DataFrame, optional
        the data frames used to fit `m1` and `m2`, only used when `sampler` is ``bootstrap``.

    Returns
    -------
    t : numpy.ndarray
        1d (or 2d, if `nsamples`>1) array of time points, in hours.
    y: numpy.ndarray
        2d (or 3d, if `nsamples`>1) array of strain frequencies.
        First axis is time, second axis is strain, third axis (if applicable) is sample.
    fig : matplotlib.figure.Figure
        figure object.
    ax : numpy.ndarray
        array of :py:class:`matplotlib.axes.Axes` objects.

    Raises
    ------
    TypeError
        if `m1` or `m2` are not :py:class:`lmfit.model.ModelResult`.
    AssertionError
        if an intermediate calculation produced an invalid result (not guaranteed).

    Examples
    --------
    >>> import pandas as pd
    >>> import curveball
    >>> plate = pd.read_csv('plate_templates/G-RG-R.csv')
    >>> df = curveball.ioutils.read_tecan_xlsx('data/Tecan_280715.xlsx', label='OD', plate=plate)
    >>> green = curveball.models.fit_model(df[df.Strain == 'G'], PLOT=False, PRINT=False)[0]
    >>> red = curveball.models.fit_model(df[df.Strain == 'R'], PLOT=False, PRINT=False)[0]
    >>> t, y = curveball.competitions.compete(green, red, PLOT=False)

    Notes
    -----
    To debug, uncomment lines to return the ``infodict`` that is returned from :py:func:`scipy.integrate.odeint` is run with ``full_output=1``.
    """
    if not isinstance(m1, lmfit.model.ModelResult):
        raise TypeError("m1 must be %s, instead it is %s", lmfit.model.ModelResult, type(m1))
    if not isinstance(m2, lmfit.model.ModelResult):
        raise TypeError("m2 must be %s, instead it is %s", lmfit.model.ModelResult, type(m2))

    if t is None:
        t = np.linspace(0, hours, num_of_points)

    if nsamples > 1:
        # draw random params
        sampler = sampler.lower()
        if sampler == 'covar':
            m1_samples = curveball.models.sample_params(m1, nsamples, params=params1)
            m2_samples = curveball.models.sample_params(m2, nsamples, params=params2)
        elif sampler == 'bootstrap':
            if params1 or params2:
                warnings.warn("Bootstrap sampling doesn't support params1 and params2 arguments")
            if df1 is None or df2 is None:
                raise ValueError("Bootstrap sampling requires kwargs df1 and df2")
            m1_fixed = {pname for pname,p in m1.params.items() if not p.vary}
            m2_fixed = {pname for pname,p in m2.params.items() if not p.vary}
            # FIXME bootstrap_params has changed
            m1_samples = curveball.models.bootstrap_params(df1, m1, nsamples,
                                                           fit_kws={'param_fix': m1_fixed})
            m2_samples = curveball.models.bootstrap_params(df2, m2, nsamples,
                                                           fit_kws={'param_fix': m2_fixed})
        else:
            raise ValueError("Unknow sampler method: {0}".format(sampler))
        min_nsamples = min(len(m1_samples), len(m2_samples))
        if nsamples > min_nsamples:
            warnings.warn("{0} resamples lost".format(nsamples - min_nsamples))
            nsamples = min_nsamples
    else:
        nsamples = 1
        # override model result params with arguments params1 and params2
        if params1:
            _params = copy.copy(m1.best_values)
            _params.update(params1)
            params1 = _params
        else:
            params1 = m1.best_values
        if params2:
            _params = copy.copy(m2.best_values)
            _params.update(params2)
            params2 = _params
        else:
            params2 = m2.best_values
        # param samples contain the model fit estimated params
        m1_samples = pd.DataFrame([params1])
        m2_samples = pd.DataFrame([params2])
        assert len(m1_samples) == len(m2_samples)

    y = np.empty((len(t), 2, nsamples))
    #infodict = [None]*nsamples # DEBUG

    # simulate the ode for each param sample
    for i in range(nsamples):
        if y0 is None:
            p0 = p0[0] / (p0[0] + p0[1]), p0[1] / (p0[0] + p0[1])
            y0 = m1_samples.iloc[i]['y0'] * p0[0], m2_samples.iloc[i]['y0'] * p0[1]
        r = m1_samples.iloc[i]['r'], m2_samples.iloc[i]['r']
        K = m1_samples.iloc[i]['K'], m2_samples.iloc[i]['K']
        nu = m1_samples.iloc[i].get('nu', 1.0), m2_samples.iloc[i].get('nu', 1.0)
        if lag_phase:
            q0 = m1_samples.iloc[i].get('q0', np.inf), m2_samples.iloc[i].get('q0', np.inf)
            v = m1_samples.iloc[i].get('v', r[0]), m2_samples.iloc[i].get('v', r[1])
        else:
            q0 = np.inf, np.inf
            v = np.inf, np.inf

        args = (r, K, nu, q0, v)
        y[:,:,i] = odeint(ode, y0=y0, t=t, args=args)

        # DEBUG
        #_y_,info = odeint(double_baranyi_roberts_ode0, y0, t, args=args, full_output=1)
        #info['args'] = (y0,) + args
        #infodict[i] = info
        #if info['message'] == 'Integration successful.':
        #   y[:,:,i] = _y_

    if PLOT:
        if ax is None:
            fig,ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        df = pd.DataFrame()
        for i in range(y.shape[1]):
            _df = pd.DataFrame(y[:,i,:])
            _df['Time'] = t
            _df = pd.melt(_df, id_vars='Time', var_name='Replicate', value_name='y')
            _df['Strain'] = i
            df = pd.concat((df, _df))

        if colors is not None:
            colors = {i:c for i,c in enumerate(colors)}
        sns.lineplot(data=df, x='Time', hue='Strain', y='y',
                        ci=ci, palette=colors, ax=ax)
        ax.set_xlabel('Time (hour)')
        ax.set_ylabel('OD')
        sns.despine()
        return t,y,fig,ax
    return t,y#,infodict


def selection_coefs_ts(t, y, ax=None, PLOT=False):
    r"""Calculate selection coefficient according to the following formula[2]_,
    where :math:`A(t), B(t)` are population densities of assay strain *A* and reference strain *B* at time *t*:

    .. math::

        s = \frac{d}{dt} \log{\frac{A(t)}{B(t)}}

    Parameters
    ----------
    t : numpy.ndarray
        array of time points, as produced by :py:func:`compete`
    y : numpy.ndarray
        array of population densities, as produced by :py:func:`compete`, where the first axis is time and the second axis is strain.
    ax : matplotlib.axes.Axes, optional
        if `PLOT` is :const:`True`, an axes to plot into; if not provided, a new one is created.
    PLOT : bool, optional
        if :const:`True`, the function will plot the curve of *s* as a function of *t*.

    Returns
    -------
    svals : numpy.ndarray
        the selection coefficients of the assay strain relative to the reference strain over time.
    fig : matplotlib.figure.Figure
        figure object.
    ax : numpy.ndarray of matplotlib.axes.Axes
        array of axes objects.

    Notes
    -----
    This formula assumes that the frequencies of the strains follow a logistic curve.
    Lag phases, interactions, etc. may cause this formula to become irrelevant.

    References
    ----------
    .. [12] Chevin, L-M. 2011. `On Measuring Selection in Experimental Evolution <http://dx.doi.org/10.1098/rsbl.2010.0580>`_. Biology Letters.
    """
    dt = np.ediff1d(t, to_end=1)
    dt[-1] = dt[-2]
    svals = np.gradient(np.log(y[:,0] / y[:,1]), dt)
    svals[np.isinf(svals)] = svals[np.isfinite(svals)].max()

    if PLOT:
        if ax is None:
            fig,ax = plt.subplots(1,1)
        else:
            fig = ax.get_figure()
        ax.plot(t, svals)
        ax.set_ylabel('Selection coefficient')
        ax.set_xlabel('Time (hour)')
        sns.despine()
        return svals, fig, ax

    return svals


def fitness_LTEE(y, ref_strain=0, assay_strain=1, t0=0, t1=-1, ci=0):
    r"""Calculate relative fitness according to the definition used in the *Long Term Evolutionary Experiment* (LTEE) [3]_,
    where :math:`A(t), B(t)` are population densities of assay strain *A* and reference strain *B* at time *t*:

    .. math::

        \omega = \frac{\log{(A(t) / A(0))}}{\log{(B(t) / B(0))}}

    If `ci` > 0, treats the third axis of `y` as replicates (from a parameteric boostrap) to calculate the confidence interval on the fitness.

    Parameters
    ----------
    y : numpy.ndarray
        array of population densities, as produced by :py:func:`compete`, where the first axis is time and the second axis is strain. A third axis is also applicable if `ci`>0.
    ref_strain : int, optional
        the index of the reference strain within `y`. This strain's fitness is set to 1 by definition. Defaults to 0 (first).
    assay_strain : int, optional
        the index of the assay strain within `y`. The result will be the fitness of this strain relative to the fitness of the reference strain. Defaults to 1 (second).
    t0 : int, optional
        the index of the time point from which to start the calculation of the relative fitness, defaults to 0 (first).
    t1 : int, optional
        the index of the time point at which to end the calculation of the relative fitness, defaults to -1 (last).
    ci : float between 0 and 1, optional
        if not zero, a confidence interval will be calculated using the third axis of `y` as replicates.

    Returns
    -------
    w : float
        the fitness of the assay strain relative to the reference strain.
    low : float
        if `ci` > 0 and `y.ndim` = 3, this is the low limit of the confidence interval of the fitness
    high : float
        if `ci` > 0 and `y.ndim` = 3, this is the higher limit of the confidence interval of the fitness

    Raises
    ------
    ValueError
        if confidence interval is requested (`ci` > 0) but there are no replicates (`y.ndim` != 3).

    Notes
    -----
    The result may depend on the choice of `t0` and `t1` as well as the strain designations (`ref_strain` and `assay_strain`).

    References
    ----------
    .. [3] Wiser, M. J. & Lenski, R. E. 2015 `A Comparison of Methods to Measure Fitness in Escherichia coli <http://dx.plos.org/10.1371/journal.pone.0126210>`_. PLoS One.
    """
    if ci == 0:
        y = y.reshape(y.shape + (1,))
    elif y.ndim != 3:
        raise ValueError()
    w = np.zeros(y.shape[2])
    for i in range(y.shape[2]):
        At0, Bt0 = y[t0, assay_strain, i], y[t0, ref_strain, i]
        At1, Bt1 = y[t1, assay_strain, i], y[t1, ref_strain, i]
        w[i] = np.log(At1 / At0) / np.log(Bt1 / Bt0)

    if ci == 0:
        return w.mean()
    else:
        margin = (1 - ci) * 50
        return w.mean(), np.percentile(w, margin), np.percentile(w, ci * 100 + margin)


def baranyi_roberts_gd(y, t, *args):
    r"""
    Fujikawa, Hiroshi, and Mohammad Z Sakha. 2014. Prediction of Microbial Growth in Mixed Culture with a Competition Model. Biocontrol Science 19 (2): 89-92.
    """
    y1, y2 = y
    K, r, nu, q0, v, a = args
    r1, r2 = r
    K1, K2 = K
    nu1, nu2 = nu
    q01, q02 = q0
    v1, v2 = v
    a1, a2 = a
    alfa1 = _alfa(t, q01, v1)
    alfa2 = _alfa(t, q02, v2)
    dy1dt = r1 * alfa1 * y1 * (1 - (y1/K1)**nu1) * (1 - (y2/K2)**(nu2*a2))
    dy2dt = r2 * alfa2 * y2 * (1 - (y1/K1)**(nu1*a1)) * (1 - (y2/K2)**nu2)
    return [dy1dt, dy2dt]


def baranyi_roberts_lv(y, t, *args):
    r"""
    Fujikawa, Hiroshi, and Mohammad Z Sakha. 2014. Prediction of Microbial Growth in Mixed Culture with a Competition Model. Biocontrol Science 19 (2): 89-92.
    """
    y1, y2 = y
    K, r, nu, q0, v, a = args
    r1, r2 = r
    K1, K2 = K
    nu1, nu2 = nu
    q01, q02 = q0
    v1, v2 = v
    a1, a2 = a
    alfa1 = _alfa(t, q01, v1)
    alfa2 = _alfa(t, q02, v2)
    Kmax = max(K1**nu1, K2**nu2)
    dy1dt = r1 * alfa1 * y1 * (1 - (y1/K1)**nu1) * (1 - (y1**a1 + y2**a2)/Kmax)
    dy2dt = r2 * alfa2 * y2 * (1 - (y2/K2)**nu2) * (1 - (y1**a1 + y2**a2)/Kmax)
    return [dy1dt, dy2dt]


def baranyi_roberts_yr2(y, t, *args):
    y1, y2 = y
    K, r, nu, q0, v, a = args
    r1, r2 = r
    K1, K2 = K
    nu1, nu2 = nu
    q01, q02 = q0
    v1, v2 = v
    a1, a2 = a
    alfa1 = _alfa(t, q01, v1)
    alfa2 = _alfa(t, q02, v2)
    dy1dt = r1 * alfa1 * y1 * (1 - (y1/K1)**nu1 - (a2 * y2/K2)**nu2)
    dy2dt = r2 * alfa2 * y2 * (1 - (a1 * y1/K1)**nu1 - (y2/K2)**nu2)
    return [dy1dt, dy2dt]


def baranyi_roberts_yr(y, t, *args):
    y1, y2 = y
    K, r, nu, q0, v, a = args
    r1, r2 = r
    K1, K2 = K
    nu1, nu2 = nu
    q01, q02 = q0
    v1, v2 = v
    a1, a2 = a
    alfa1 = _alfa(t, q01, v1)
    alfa2 = _alfa(t, q02, v2)
    dy1dt = r1 * alfa1 * y1 * (1 - (y1**nu1) / (K1**nu1) - a2 * (y2**nu2) / (K1**nu1))
    dy2dt = r2 * alfa2 * y2 * (1 - a1 * (y1**nu1) / (K2**nu2) - (y2**nu2) / (K2**nu2))
    return [dy1dt, dy2dt]

# obsolete - just call baranyi_roberts_yr(y, t, *args, (1, 1))
# def baranyi_roberts_yr_a1(y, t, *args):
#     return baranyi_roberts_yr(y, t, *args, (1, 1))


def fit_and_compete(m1, m2, df_mixed, y0=None, aguess=(1, 1), fixed=False,
                    ode=baranyi_roberts_yr, num_of_points=100, method='nelder',
                    value_var = 'OD', time_var = 'Time',
                    PLOT=False, colors=sns.color_palette('Set1', 3)):
    best_values1 = m1.best_values if hasattr(m1, 'best_values') else m1
    best_values2 = m2.best_values if hasattr(m2, 'best_values') else m2
    K = best_values1['K'], best_values2['K']
    r = best_values1['r'], best_values2['r']
    nu = best_values1.get('nu',1), best_values2.get('nu',1)
    q0 = best_values1.get('q0', np.inf), best_values2.get('q0', np.inf)
    v = best_values1.get('v', r[0]), best_values2.get('v', r[1])
    if y0 is None:
        y0 = best_values1['y0']/2, best_values2['y0']/2

    y_mixed = df_mixed.groupby(time_var)[value_var].mean().values
    y_mixed_std = df_mixed.groupby(time_var)[value_var].std().values
    t_mixed = np.unique(df_mixed[time_var])
    t = np.linspace(0, t_mixed.max(), num_of_points)

    if fixed:
        a = aguess
        a1, a2 = a
    else:
        def mixed_model(t, a1, a2):
            y = odeint(ode, y0, t, args=(K, r, nu, q0, v, (a1, a2)))
            return y.sum(axis=1)
        model = lmfit.Model(mixed_model)

        params = model.make_params(a1=aguess[0], a2=aguess[1])
        params['a1'].set(min=1e-1, max=10)
        params['a2'].set(min=1e-1, max=10)
        
        result = model.fit(data=y_mixed, t=t_mixed, params=params, weights=1.0 / y_mixed_std, 
            method=method, nan_policy='propagate')

        a = result.best_values['a1'], result.best_values['a2']
        a1, a2 = a

    y = odeint(ode, y0, t, args=(K, r, nu, q0, v, a))

    if PLOT:
        ysum = y.sum(axis=1)
        p1 = y[:, 0] / ysum
        p2 = y[:, 1] / ysum
        MRSE = ((odeint(ode, y0, t_mixed, args=(K, r, nu, q0, v, a)).sum(axis=1) - y_mixed)**2).mean()

        w, h = plt.rcParams['figure.figsize']
        fig, ax = plt.subplots(1, 2, figsize=(w * 2, h), sharex=True)


        ax[0].plot(t, y[:, 0], color=colors[0])
        ax[0].plot(t, y[:, 1], color=colors[1])
        ax[0].plot(t, ysum, color=colors[2])
        if not fixed:
            ax[0].plot(t_mixed, result.init_fit, ls='-.', color=colors[2])
        ax[0].errorbar(t_mixed, y_mixed, y_mixed_std, fmt='o', color=colors[2])

        ax[0].set(xlabel='Time (hour)', ylabel='OD',
                  xlim=(-0.5, t.max() + 0.5),
                  title="MRSE: {:.2g}".format(MRSE))

        ax[1].plot(t, p1, color=colors[0])
        ax[1].plot(t, p2, color=colors[1])

        ax[1].set(xlabel='Time (hour)', ylabel='Frequency',
                  xlim=(-0.5, t.max() + 0.5), ylim=(0, 1),
                  title="a1={:.2g}, a2={:.2g}".format(*a))

        sns.despine()
        fig.tight_layout()
        return t, y, a, fig, ax
    return t, y, a


def fit_and_compete_ci(param_samples1, param_samples2, df_mixed, y0=None, ci=0.95,
    colors=('g', 'r'), line_kws=None, ci_kws=None, ax=None, PLOT=False):
    if line_kws is None:
        line_kws = dict()
    if ci_kws is None:
        ci_kws = dict(color='gray', alpha=0.5)
    nsamples = param_samples1.shape[0]
    assert param_samples1.shape[0] == param_samples2.shape[0], "Parameters samples should have the same length"
    t = [None] * nsamples
    y = [None] * nsamples
    a = [None] * nsamples
    for i in range(nsamples):
        sample1 = param_samples1.iloc[i, :]
        sample2 = param_samples2.iloc[i, :]
        t[i], y[i], a[i] = curveball.competitions.fit_and_compete(sample1, sample2, df_mixed, y0=y0)

    t = np.array(t)
    y = np.array(y)
    a = np.array(a)
    ysum = y.sum(axis=2)
    f1 = y[:, :, 0] / ysum
    f2 = y[:, :, 1] / ysum
    f = np.array((f1, f2)).T

    margin = (1.0 - ci) * 50.0

    low_f = np.percentile(f, margin, axis=1)
    high_f = np.percentile(f, ci * 100.0 + margin, axis=1)
    avg_f = f.mean(axis=1)

    low_a = np.percentile(a, margin, axis=0)
    avg_a = a.mean(axis=0)
    high_a = np.percentile(a, ci * 100.0 + margin, axis=0)

    if PLOT:
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        ax.fill_between(t[0, :], low_f[:, 0], high_f[:, 0], **ci_kws)
        ax.plot(t[0,:], avg_f[:, 0], color=colors[0], **line_kws)

        ax.fill_between(t[0, :], low_f[:, 1], high_f[:, 1], **ci_kws)
        ax.plot(t[0,:], avg_f[:, 1], color=colors[1], **line_kws)

        ax.set(xlabel='Time', ylabel='Frequency')
        fig.tight_layout()
        sns.despine()
        return low_a, avg_a, high_a, low_f, avg_f, high_f, fig, ax

    return low_a, avg_a, high_a, low_f, avg_f, high_f


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import print_function
from __future__ import division
from builtins import str

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import sklearn.linear_model
import sklearn.preprocessing
import sklearn.pipeline
try:
    from sklearn.model_selection import GridSearchCV
except ImportError:
    from sklearn.grid_search import GridSearchCV
import webcolors


def smooth(x, y, PLOT=False, **kwargs):
    """Estimate a smoothing function.

    The function finds a polynomial function that fits to the data using a linear regression model
    with polynomial features and using a cross-validation grid search to find the best polynomial degree.

    Parameters
    ----------
    x : numpy.ndarray
        array of floats for the independent variable
    y : numpy.ndarray
        array of floats for the dependent variable
    PLOT : bool, optional
        if :const:`True`, plots a figure of the input and smoothed data, defaults to :const:`False`
    kwargs : optional
        extra keyword arguments passed to the underlying smoothing function.

    Returns
    -------
    f : function
        smooth function that corresponds to the data.
    fig : matplotlib.figure.Figure
        if the argument `PLOT` was :const:`True`, the generated figure.
    ax : matplotlib.axes.Axes
        if the argument `PLOT` was :const:`True`, the generated axis.
    """
    assert len(x) == len(y)
    x = np.array(x)
    y = np.array(y)

    # do a grid search to find the optimal polynomial degree
    model = GridSearchCV(
        # use a linear model with polynomial features
        sklearn.pipeline.Pipeline([
            ('poly', sklearn.preprocessing.PolynomialFeatures(degree=3)),
            ('linear', sklearn.linear_model.LinearRegression())
        ]),
        cv=kwargs.get('cv', min(5, len(x))),
        param_grid={
            'poly__degree': np.arange(
                kwargs.get('min_degree', 3),
                kwargs.get('max_degree', 14),
                2
            )
        }
    )
    x = x.reshape(-1, 1)
    model.fit(x, y)

    def predict(_x):
        return model.predict(np.array(_x).reshape(-1, 1))

    if PLOT:
        yhat = predict(x)
        fig, ax = plt.subplots(1, 1)
        ax.plot(x, yhat, 'k--')
        ax.plot(x, y, 'ko')
        ax.set(xlabel='x', ylabel='y')
        sns.despine()
        return predict, fig, ax
    if kwargs.get('PRINT'):
        print(model)
    return predict


def color_name_to_hex(name, default='#000000'):
    """Convert a color name to a hex values.

    Does not attempt to convert names that already start with '#'.

    Parameters
    ----------
    name : str
        color name to convert, i.e. "red"
    default : str
        the default color in case name is invalid, default value is "#000000".

    Returns
    -------
    str
        hex value corresponding to name
    """
    try:
        name = str(name)
        if name.startswith('#'):
            return name
        return webcolors.name_to_hex(name)
    except ValueError:
        return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
import matplotlib
matplotlib.use('Agg')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from unittest import TestCase, main
from importlib import resources
import shutil
import tempfile
import os
import pandas as pd
import numpy as np
from PIL import Image
import seaborn as sns
import curveball
import lmfit
import matplotlib
import logging
logging.getLogger('matplotlib').setLevel(logging.INFO)


CI = os.environ.get('CI', 'false').lower() == 'true'


class MplTestCase(TestCase):
	def test_mpl(self):
		if CI:
			self.assertEqual(matplotlib.rcParams['backend'].lower(), 'agg')


class PlotsTestCase(TestCase):
	def setUp(self):
		self.folder = tempfile.mkdtemp()
		self.output_filename = os.path.join(self.folder, 'plate.png')


	def tearDown(self):
		shutil.rmtree(self.folder)


	def check_image(self):		
		im = Image.open(self.output_filename)
		bands = im.split()
		self.assertFalse(all(band.getextrema() == (255, 255) for band in bands))
		

	def test_plot_wells(self):
		df = pd.read_csv(resources.files("data").joinpath("Tecan_210115.csv"))
		g = curveball.plots.plot_wells(df, output_filename=self.output_filename)
		self.assertIsInstance(g, sns.axisgrid.Grid)
		self.check_image()


	def test_plot_strains(self):
		df = pd.read_csv(resources.files("data").joinpath("Tecan_210115.csv"))
		g = curveball.plots.plot_strains(df, output_filename=self.output_filename)
		self.assertIsInstance(g, sns.axisgrid.Grid)
		self.check_image()


	def test_plot_strains_no_output(self):
		df = pd.read_csv(resources.files("data").joinpath("Tecan_210115.csv"))
		g = curveball.plots.plot_strains(df)
		self.assertIsInstance(g, sns.axisgrid.Grid)
		with self.assertRaises(IOError):
			self.check_image()


	def test_tsplot(self):
		df = pd.read_csv(resources.files("data").joinpath("Tecan_210115.csv"))		
		g = curveball.plots.tsplot(df, output_filename=self.output_filename)
		self.assertIsInstance(g, matplotlib.axes.Axes)
		self.check_image()


	def test_plot_plate(self):
		df = pd.read_csv(resources.files("plate_templates").joinpath("checkerboard.csv"))		
		fig, ax = curveball.plots.plot_plate(df, output_filename=self.output_filename)
		self.assertIsInstance(fig, matplotlib.figure.Figure)
		self.check_image()


	def test_plot_params_distribution(self):
		df = pd.DataFrame(np.random.normal(0, 1, size=(100, 4)), columns=list('ABCD'))
		g = curveball.plots.plot_params_distribution(df)
		self.assertIsInstance(g, sns.axisgrid.Grid)

	
	def test_plot_residuals(self):
		def f(t, a, b, c):
			return a + b * t + c * t**2
		t = np.linspace(0, 100)
		a, b, c = 1, 2, 3		
		y = f(t, a, b, c) + np.random.normal(0, 0.01)
		data = pd.DataFrame({'Time': t, 'OD': y})
		fig, ax = curveball.plots.plot_residuals(data)
		self.assertIsInstance(fig, matplotlib.figure.Figure)	


	def test_plot_model_residuals(self):
		def f(t, a, b, c):
			return a + b * t + c * t**2
		t = np.linspace(0, 100)
		a, b, c = 1, 2, 3		
		y = f(t, a, b, c) + np.random.normal(0, 0.01)
		model = lmfit.model.Model(f)
		params = model.make_params(a=1, b=1, c=1)
		result = model.fit(data=y, t=t, params=params)
		fig, ax = curveball.plots.plot_model_residuals(result)		
		self.assertIsInstance(fig, matplotlib.figure.Figure)	


if __name__ == '__main__':
	main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import str
from builtins import filter
from builtins import map
from builtins import range
from six import string_types
from past.utils import old_div
import numpy as np
import pandas as pd
from string import ascii_uppercase
from scipy.io import loadmat
import re
import datetime
import dateutil.parser
from glob import glob
import os
import os.path
from warnings import warn


MAT_VERSION = u'1.0'


def _fix_dtypes(df):
    """Fix the dtypes of categorial columns.

    Parameters
    ----------
    df : pandas.DataFrame
        data frame to fix

    Returns
    -------
    None
    """
    if u'Strain' in df.columns:
        df[u'Strain'] = df[u'Strain'].astype(str)
    for col in df.columns:
        if df[col].dtype == object:
            df[col] = df[col].astype(str)
    

def read_curveball_csv(filename, max_time=None, plate=None):
    """Reads growth measurements from a Curveball csv (comma separated values) file.

    Parameters
    ----------
    filename : str
        path to the file.
    plate : pandas.DataFrame, optional
        data frame representing a plate, usually generated by reading a CSV file generated by `Plato <http://plato.yoavram.com/>`_.

    Returns
    -------
    pandas.DataFrame

    Examples
    --------   
    >>> df = curveball.ioutils.read_curveball_csv("data/Tecan_210115.csv")
    """
    df = pd.read_csv(filename, encoding='utf8')
    if plate is None and 'Strain' not in df.columns:
        df[u'Strain'] = u'0'
    if plate is None and 'Color' not in df.columns:
        df[u'Color'] = u'#000000'
    if max_time is not None:
        df = df[df.Time <= max_time].copy()
    _fix_dtypes(df)
    return df


def write_curveball_csv(df, filename):
    """Reads growth measurements from a Curveball csv (comma separated values) file.

    Parameters
    ----------
    df : pandas.DataFrame, optional
        data frame to write
    filename : str
        path to the output file        
    """
    df.to_csv(filename, index=False, encoding='utf8')


def read_tecan_xlsx(filename, label=u'OD', sheets=None, max_time=None, plate=None, PRINT=False):
    """Reads growth measurements from a Tecan Infinity Excel output file.

    Parameters
    ----------
    filename : str
        path to the file.
    label : str / sequence of str
        a string or sequence of strings containing measurment names used as titles of the data tables in the file.
    sheets : list, optional
        list of sheet numbers, if known. Otherwise the function will try to all the sheets.
    max_time : float, optional
        maximal time in hours, defaults to infinity
    plate : pandas.DataFrame, optional
        data frame representing a plate, usually generated by reading a CSV file generated by `Plato <http://plato.yoavram.com/>`_.

    Returns
    -------
    pandas.DataFrame
        Data frame containing the columns:

        - ``Time`` (:py:class:`float`, in hours)
        - ``Temp. [°C]`` (:py:class:`float`)
        - ``Cycle Nr.`` (:py:class:`int`)
        - ``Well`` (:py:class:`str`): the well name, usually a letter for the row and a number of the column.
        - ``Row`` (:py:class:`str`): the letter corresponding to the well row.
        - ``Col`` (:py:class:`str`): the number corresponding to the well column.
        - ``Strain`` (:py:class:`str`): if a `plate` was given, this is the strain name corresponding to the well from the plate.
        - ``Color`` (:py:class:`str`, hex format): if a `plate` was given, this is the strain color corresponding to the well from the plate.

        There will also be a separate column for each label, and if there is more than one label, a separate `Time` and `Temp. [°C]` column for each label.

    Raises
    ------
    ValueError
        if not data was parsed from the file.

    Examples
    --------
    
    >>> plate = pd.read_csv("plate_templates/G-RG-R.csv")
    >>> df = curveball.ioutils.read_tecan_xlsx("data/Tecan_210115.xlsx", label=('OD','Green','Red'), max_time=12, plate=plate)
    >>> df.shape
    (8544, 9)
    """
    import numbers
    import openpyxl
    from openpyxl.utils.datetime import from_excel
    wb = openpyxl.load_workbook(filename, data_only=True, read_only=True)
    dateandtime = datetime.datetime.now() # default

    if isinstance(label, string_types):
        label = [label]
    if sheets is None:
        sheets = range(len(wb.worksheets))
    if PRINT: print("Reading {0} worksheets from workbook {1}".format(len(sheets), filename))
    label_dataframes = []
    for lbl in label:
        sheet_dataframes = []        
        ## FOR sheet
        for sh_i in sheets:
            sh = wb.worksheets[sh_i]
            rows = list(sh.iter_rows(values_only=True))
            if not rows:
                continue # to next sheet
        
            for i in range(len(rows)):
                ## FOR row
                row = rows[i]
                first = row[0]
                if isinstance(first, string_types) and first.startswith(u'Date'):
                    if isinstance(row[1], string_types):
                        date = ''.join([x for x in row[1:] if x])
                        next_row = rows[i + 1]
                        if isinstance(next_row[0], string_types) and next_row[0].startswith(u'Time'):
                            time = ''.join([x for x in next_row[1:] if x])
                        else:
                            warn(u"Warning: time row missing (sheet '{0}', row{1}), found row starting with {2}".format(sh.name, i, row[0]))
                        dateandtime = dateutil.parser.parse("%s %s" % (date, time))
                    elif isinstance(row[1], datetime.datetime):
                        date_dt = row[1]
                        next_row = rows[i + 1]
                        if isinstance(next_row[0], string_types) and next_row[0].startswith(u'Time'):
                            time_val = next_row[1]
                            if isinstance(time_val, string_types):
                                time = dateutil.parser.parse(time_val).time()
                            elif isinstance(time_val, datetime.time):
                                time = time_val
                            elif isinstance(time_val, numbers.Number):
                                time = from_excel(time_val, wb.epoch).time()
                            else:
                                time = date_dt.time()
                            dateandtime = datetime.datetime.combine(date_dt.date(), time)
                        else:
                            warn(u"Warning: time row missing (sheet '{0}', row{1}), found row starting with {2}".format(sh.name, i, row[0]))
                            dateandtime = date_dt
                    elif isinstance(row[1], datetime.date):
                        date_dt = datetime.datetime.combine(row[1], datetime.time())
                        next_row = rows[i + 1]
                        if isinstance(next_row[0], string_types) and next_row[0].startswith(u'Time'):
                            time_val = next_row[1]
                            if isinstance(time_val, string_types):
                                time = dateutil.parser.parse(time_val).time()
                            elif isinstance(time_val, datetime.time):
                                time = time_val
                            elif isinstance(time_val, numbers.Number):
                                time = from_excel(time_val, wb.epoch).time()
                            else:
                                time = datetime.time()
                            dateandtime = datetime.datetime.combine(date_dt.date(), time)
                        else:
                            warn(u"Warning: time row missing (sheet '{0}', row{1}), found row starting with {2}".format(sh.name, i, row[0]))
                            dateandtime = date_dt
                    elif isinstance(row[1], numbers.Number):
                        date_dt = from_excel(row[1], wb.epoch)
                        next_row = rows[i + 1]
                        if isinstance(next_row[0], string_types) and next_row[0].startswith(u'Time'):
                            time_val = next_row[1]
                            if isinstance(time_val, string_types):
                                time = dateutil.parser.parse(time_val).time()
                            elif isinstance(time_val, datetime.time):
                                time = time_val
                            elif isinstance(time_val, numbers.Number):
                                time = from_excel(time_val, wb.epoch).time()
                            else:
                                time = date_dt.time()
                            dateandtime = datetime.datetime.combine(date_dt.date(), time)
                        else:
                            warn(u"Warning: time row missing (sheet '{0}', row{1}), found row starting with {2}".format(sh.name, i, row[0]))
                            dateandtime = date_dt
                    else:
                        warn(u"Warning: date row (sheet '{2}', row {3}) could not be parsed: {0} {1}".format(row[1], type(row[1]), sh.name, i))
                if first == lbl:
                    break
                ## FOR row ENDS
            
            data = {}            
            for j in range(i + 1, len(rows)):
                ## FOR row
                row = rows[j]
                if not row[0]:
                    break
                data[row[0]] = [x for x in row[1:] if isinstance(x, numbers.Number)]
                ## FOR row ENDS

            if not data:
                continue

            min_length = min(map(len, data.values()))
            for k,v in data.items():
                data[k] =  v[:min_length]

            df = pd.DataFrame(data)
            df = pd.melt(df, id_vars=(u'Time [s]', u'Temp. [°C]', u'Cycle Nr.'), var_name=u'Well', value_name=lbl)
            df.rename(columns={u'Time [s]': u'Time'}, inplace=True)
            df.Time = [dateandtime + datetime.timedelta(0, t) for t in df.Time]        
            df[u'Row'] = [x[0] for x in df.Well]
            df[u'Col'] = [int(x[1:]) for x in df.Well]
            sheet_dataframes.append(df)
        ## FOR sheet ENDS

        n_sheet_dataframes = len(sheet_dataframes)
        if n_sheet_dataframes == 0:
            raise ValueError("No data found in any sheet of workbook {0}".format(filename))
        elif n_sheet_dataframes == 1:
            df = sheet_dataframes[0]
        else:
            df = pd.concat(sheet_dataframes)

        min_time = df.Time.min()
        if PRINT:
            print("Starting time", min_time)
        df.Time = [(t - min_time).total_seconds() / 3600.0 for t in df.Time]
        if max_time is not None:
            df = df[df.Time <= max_time].copy()
        df = df.sort_values([u'Row', u'Col', u'Time'])
        label_dataframes.append((lbl,df))

    n_label_dataframes = len(label_dataframes)
    if n_label_dataframes == 0: # no dataframes
        return pd.DataFrame()
    if n_label_dataframes == 1: # just one dataframe
        df = label_dataframes[0][1]
    else: # multiple dataframes, merge together
        # FIXME last label isn't used as a suffix, not sure why
        lbl, df = label_dataframes[0]
        lbl = '_' + lbl
        for lbli, dfi in label_dataframes[1:]:
            lbli = '_' + lbli
            df = pd.merge(df, dfi, on=(u'Cycle Nr.', u'Well', u'Row', u'Col'), suffixes=(lbl,lbli))
    if plate is None:
        df[u'Strain'] = u'0'
        df[u'Color'] = u'#000000'
    else:
        df = pd.merge(df, plate, on=(u'Row', u'Col'))
    if PRINT: print("Read {0} records from workbook".format(df.shape[0]))
    _fix_dtypes(df)
    return df


def read_tecan_mat(filename, time_label=u'tps', value_label=u'plate_mat', value_name=u'OD', plate_width=12, max_time=None, plate=None):
    """Reads growth measurements from a Matlab file generated by a propriety script at the *Pilpel lab*.

    Parameters
    ----------
    filename : str
        name of the XML file to be read. Use ``*`` and ``?`` in filename to read multiple files and parse them into a single data frame.
    time_label : str, optional
        name of the field used to store the time values, defaults to ``tps``.
    label : str
        name of the field used to store the OD values, defaults to ``plate_mat``.
    plate_width : int
        width of the microwell in plate in number of wells, defaults to 12.
    max_time : float, optional
        maximal time in hours, defaults to infinity
    plate : pandas.DataFrame, optional
        data frame representing a plate, usually generated by reading a CSV file generated by `Plato <http://plato.yoavram.com/>`_.

    Returns
    -------
    pandas.DataFrame
        Data frame containing the columns:

        - ``Time`` (:py:class:`float`, in hours)
        - ``OD`` (:py:class:`float`)
        - ``Well`` (:py:class:`str`): the well name, usually a letter for the row and a number of the column.
        - ``Row`` (:py:class:`str`): the letter corresponding to the well row.
        - ``Col`` (:py:class:`str`): the number corresponding to the well column.
        - ``Filename`` (:py:class:`str`): the filename from which this measurement was read.
        - ``Strain`` (:py:class:`str`): if a `plate` was given, this is the strain name corresponding to the well from the plate.
        - ``Color`` (:py:class:`str`, hex format): if a `plate` was given, this is the strain color corresponding to the well from the plate.
    """
    mat = loadmat(filename, appendmat=True)
    if mat[u'__version__'] != MAT_VERSION:
        warn(u"Warning: expected mat file version {0} but got {1}".format(MAT_VERSION, mat[u'__version__']))
    t = mat[time_label]
    t = t.reshape(max(t.shape))
    y = mat[value_label]
    assert y.shape[1] == t.shape[0]

    df = pd.DataFrame(y.T, columns=np.arange(y.shape[0]) + 1)
    df[u'Time'] = old_div(t, 3600.)
    df[u'Cycle Nr.'] = np.arange(1, 1 + len(t))
    df = pd.melt(df, id_vars=(u'Cycle Nr.', u'Time'), var_name=u'Well', value_name=value_name)
    df[u'Well'] = [ascii_uppercase[old_div((int(w) - 1), plate_width)] + str(w % plate_width if w % plate_width > 0 else plate_width) for w in df[u'Well']]
    df[u'Row'] = [w[0] for w in df[u'Well']]
    df[u'Col'] = [int(w[1:]) for w in df[u'Well']]
    
    if plate is None:
        df[u'Strain'] = u'0'
        df[u'Color'] = u'#000000'
    else:
        df = pd.merge(df, plate, on=(u'Row', u'Col'))
    if not max_time:
        max_time = df.Time.max()
    df = df[df.Time < max_time].copy()
    df.sort_values([u'Row', u'Col', u'Time'], inplace=True)    
    _fix_dtypes(df)
    return df


def read_tecan_xml(filename, label=u'OD', max_time=None, plate=None):
    """Reads growth measurements from a Tecan Infinity XML output files.

    Parameters
    ----------
    filename : str
        pattern of the XML files to be read. Use ``*`` and ``?`` in filename to read multiple files and parse them into a single data frame.
    label : str, optional
        measurment name used as ``Name`` in the measurement sections in the file, defaults to ``OD``.
    max_time : float, optional
        maximal time in hours, defaults to infinity
    plate : pandas.DataFrame, optional
        data frame representing a plate, usually generated by reading a CSV file generated by `Plato <http://plato.yoavram.com/>`_.

    Returns
    -------
    pandas.DataFrame
        Data frame containing the columns:

        - ``Time`` (:py:class:`float`, in hours)
        - ``Well`` (:py:class:`str`): the well name, usually a letter for the row and a number of the column.
        - ``Row`` (:py:class:`str`): the letter corresponding to the well row.
        - ``Col`` (:py:class:`str`): the number corresponding to the well column.
        - ``Filename`` (:py:class:`str`): the filename from which this measurement was read.
        - ``Strain`` (:py:class:`str`): if a `plate` was given, this is the strain name corresponding to the well from the plate.
        - ``Color`` (:py:class:`str`, hex format): if a `plate` was given, this is the strain color corresponding to the well from the plate.

    There will also be a separate column for the value of the label.

    Examples
    --------    

    >>> import zipfile
    >>> with zipfile.ZipFile("data/20130211_dh.zip") as z:
        z.extractall("data/20130211_dh")
    >>> plate = pd.read_csv("plate_templates/checkerboard.csv")
    >>> df = curveball.ioutils.read_tecan_xlsx("data/20130211_dh/*.xml", 'OD', plate=plate)
    >>> df.shape
    (2016, 8)

    Notes
    -----
    This function was adapted from `choderalab/assaytools <https://github.com/choderalab/assaytools/blob/908471e7976e207df3f9b0e31b2a89f84da40607/AssayTools/platereader.py>`_ (licensed under LGPL).
    """
    from lxml import etree
    dataframes = []
    for filename in glob(filename):
        # Parse XML file into nodes.
        root_node = etree.parse(filename)

        # Build a dict of section nodes.
        section_nodes = { section_node.get(u'Name') : section_node for section_node in root_node.xpath(u"/*/Section") }

        # Process all sections.
        if label not in section_nodes:
            return pd.DataFrame()

        section_node = section_nodes[label]
        
        # Get the time of measurement
        time_start = section_node.attrib[u'Time_Start']

        # Get a list of all well nodes
        well_nodes = section_node.xpath(u"*/Well")
        
        # Process all wells into data.
        well_data = []
        for well_node in well_nodes:
            well = well_node.get(u'Pos')
            value = float(well_node.xpath(u"string()"))
            well_data.append({u'Well': well, label: value})

        # Add to data frame
        df = pd.DataFrame(well_data)
        df[u'Row'] = [x[0] for x in df.Well]
        df[u'Col'] = [int(x[1:]) for x in df.Well]
        df[u'Time'] = dateutil.parser.parse(time_start)
        df[u'Filename'] = os.path.split(filename)[-1]
        dataframes.append(df)
    df = pd.concat(dataframes)
    min_time = df.Time.min()
    df.Time = [(t - min_time).total_seconds() / 3600.0 for t in df.Time]
    if plate is None:
        df[u'Strain'] = u'0'
        df[u'Color'] = u'#000000'
    else:
        df = pd.merge(df, plate, on=(u'Row', u'Col'))
    if max_time is not None:
        df = df[df.Time <= max_time].copy()
    df.sort_values([u'Row', u'Col', u'Time'], inplace=True)
    _fix_dtypes(df)
    return df


def read_sunrise_xlsx(filename, label=u'OD', max_time=None, plate=None):
    """Reads growth measurements from a Tecan Sunrise Excel output file.

    Parameters
    ----------
    filename : str
        pattern of the XLSX files to be read. Use * and ? in filename to read multiple files and parse them into a single data frame.    label : str, optional
    label : str, optional
        measurment name to use for the data in the file, defaults to ``OD``.
    max_time : float, optional
        maximal time in hours, defaults to infinity
    plate : pandas.DataFrame, optional
        data frame representing a plate, usually generated by reading a CSV file generated by `Plato <http://plato.yoavram.com/>`_.

    Returns
    -------
    pandas.DataFrame
        Data frame containing the columns:

        - ``Time`` (:py:class:`float`, in hours)
        - ``OD`` (or the value of `label`, if given)
        - ``Well`` (:py:class:`str`): the well name, usually a letter for the row and a number of the column.
        - ``Row`` (:py:class:`str`): the letter corresponding to the well row.
        - ``Col`` (:py:class:`str`): the number corresponding to the well column.
        - ``Filename`` (:py:class:`str`): the filename from which this measurement was read.
        - ``Strain`` (:py:class:`str`): if a `plate` was given, this is the strain name corresponding to the well from the plate.
        - ``Color`` (:py:class:`str`, hex format): if a `plate` was given, this is the strain color corresponding to the well from the plate.
    """
    import numbers
    import openpyxl
    from openpyxl.utils.datetime import from_excel
    dataframes = []
    filename = os.fspath(filename)
    files = glob(filename)
    if not files:
        return pd.DataFrame()
    for filename in files:
        wb = openpyxl.load_workbook(filename, data_only=True, read_only=True)
        for sh in wb.worksheets:
            rows = list(sh.iter_rows(values_only=True))
            if rows:
                break
        parse_data = False # start with metadata
        index = []
        data = []

        for i in range(len(rows)):
            row = rows[i]
            if row[0] == u'Date:':
                date_val = next(filter(lambda x: x is not None, row[1:]))
                if isinstance(date_val, datetime.datetime):
                    date_dt = date_val
                elif isinstance(date_val, datetime.date):
                    date_dt = datetime.datetime.combine(date_val, datetime.time())
                elif isinstance(date_val, numbers.Number):
                    date_dt = from_excel(date_val, wb.epoch)
                else:
                    raise ValueError("Unexpected date value in file {}".format(filename))
            elif row[0] == u'Time:':
                time_val = next(filter(lambda x: x is not None, row[1:]))
                if isinstance(time_val, datetime.datetime):
                    time_dt = time_val.time()
                elif isinstance(time_val, datetime.time):
                    time_dt = time_val
                elif isinstance(time_val, numbers.Number):
                    time_dt = from_excel(time_val, wb.epoch).time()
                else:
                    raise ValueError("Unexpected time value in file {}".format(filename))
            elif row[0] == u'<>':
                columns = list(map(int, filter(lambda x: x is not None, row[1:])))
                parse_data = True
            elif (row[0] is None or row[0] == '') and parse_data:
                break
            elif parse_data:
                index.append(row[0])
                data.append(list(map(float, filter(lambda x: x is not None, row[1:]))))
                
        dateandtime = datetime.datetime.combine(date_dt.date(), time_dt)
        
        df = pd.DataFrame(data, columns=columns, index=index)
        df[u'Row'] = index
        df = pd.melt(df, id_vars=u'Row', var_name=u'Col', value_name=label)
        df[u'Time'] = dateandtime
        df[u'Well'] = [x[0] + str(x[1]) for x in zip(df.Row, df.Col)]
        df[u'Filename'] = os.path.split(filename)[-1]
        dataframes.append(df)
    df = pd.concat(dataframes)
    min_time = df.Time.min()
    df.Time = [(t - min_time).total_seconds() / 3600.0 for t in df.Time]
    if plate is None:
        df[u'Strain'] = u'0'
        df[u'Color'] = u'#000000'
    else:
        df = pd.merge(df, plate, on=(u'Row', u'Col'))
    if max_time is not None:
        df = df[df.Time <= max_time].copy()
    df.sort_values([u'Row', u'Col', u'Time'], inplace=True)
    _fix_dtypes(df)
    return df


def read_biotek_xlsx(filename, max_time=None, plate=None, PRINT=False):
    import openpyxl
    wb = openpyxl.load_workbook(filename, data_only=True, read_only=True)
    for sh_i in range(len(wb.worksheets)):
        sh = wb.worksheets[sh_i]
        rows = list(sh.iter_rows(values_only=True))
        if rows:
            break
    else: # all sheets are empty
        warnings.warn('All sheets are empty in {}'.format(filename))
        return pd.DataFrame()

    if PRINT:
        print("Reading worksheet {0} with {2} lines from workbook {1}".format(sh_i, filename, len(rows)))

    in_data = False
    data = []
    dfs = []
    i = 0
    while i < len(rows):
        row = rows[i]
        if not in_data and row[1] == 'Time':
            columns = row[1:]
            in_data = True
        elif in_data and row[1] not in (None, ''):
            row = list(row)
            if isinstance(row[1], datetime.time):
                row[1] = row[1].hour + row[1].minute / 60.0 + row[1].second / 3600.0
            elif isinstance(row[1], datetime.timedelta):
                row[1] = row[1].total_seconds() / 3600.0
            elif isinstance(row[1], datetime.datetime):
                row[1] = row[1].hour + row[1].minute / 60.0 + row[1].second / 3600.0
            else:
                row[1] *= 24 # days -> hours
            data.append(row[1:])
        elif in_data and row[1] in (None, ''):
            in_data = False
            df = pd.DataFrame(data, columns=columns)
            df = pd.melt(df, [u'Time', u'T° 600'], var_name=u'Well', value_name=u'OD')
            dfs.append(df)
        i += 1
    df = pd.concat(dfs )

    df[u'Row'] = [x[0] for x in df.Well]
    df[u'Col'] = [int(x[1:]) for x in df.Well]

    min_time = df.Time.min()
    if PRINT:
        print("Starting time", min_time)
    if max_time is not None:
        df = df[df.Time <= max_time].copy()
    df.sort_values([u'Row', u'Col', u'Time'], inplace=True)

    if df.shape[0] == 0: # no dataframes
        return pd.DataFrame()

    if plate is None:
        df[u'Strain'] = u'0'
        df[u'Color'] = u'#000000'
    else:
        df = pd.merge(df, plate, on=(u'Row', u'Col'))
    if PRINT: print("Read {0} records from workbook".format(df.shape[0]))
    _fix_dtypes(df)
    return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
import sys
import numbers
from warnings import warn
import numpy as np
import matplotlib.pyplot as plt
import collections
try:
	from scipy.stats.distributions import chi2
	chisqprob = chi2.sf
except ImportError:
	from scipy.stats import chisqprob
from scipy.stats import linregress
from scipy.misc import derivative
import pandas as pd
import copy
import inspect
import lmfit
import sympy
import seaborn as sns
sns.set_style("ticks")
import curveball
import curveball.baranyi_roberts_model
from curveball.utils import smooth


def is_model(cls):
    """Returns :const:`True` if the input is a subclass of :py:class:`lmfit.model.Model`.

    Parameters
    ----------
    cls : class

    Returns
    -------
    bool
    """
    return inspect.isclass(cls) and issubclass(cls, lmfit.model.Model)


def get_models(module):
    """Finds and returns all models in `module`.

    Parameters
    ----------
    module : module
        the module in which to look for models

    Returns
    -------
    list
        list of subclasses of :py:class:`lmfit.model.Model` that can be used with :py:func:`curveball.models.fit_model`

    See also
    --------
    fit_model
    is_model
    curveball.baranyi_roberts_model
    """
    return [m[1] for m in inspect.getmembers(module, curveball.models.is_model)]


def bootstrap_params(df, model_result, nsamples, unit='Well', fit_kws=None):
    """Sample model parameters by fitting the model to resampled data.

    The data is bootstraped by drawing growth curves (sampling from the `unit` column in `df`) with replacement.

    Parameters
    ----------
    df : pandas.DataFrame
        the data to fit
    model_result : lmfit.model.ModelResult
        result of fitting a :py:class:`lmfit.model.Model` to data in ``df``
    nsamples : int
        number of samples to draw
    unit : str, optional
        the name of the column in `df` that identifies a resampling unit, defaults to ``Well``
    fit_kws : dict, optional
        dict of kwargs for `fit_model`
    Returns
    -------
    pandas.DataFrame
        data frame of samples; each row is a sample, each column is a parameter.

    Raises
    ------
    ValueError : if `model_result` isn't an instance of :py:class:`lmfit.model.ModelResult`
    ValueError : if `df` is empty

    See also
    --------
    sample_params
    """
    if not isinstance(model_result, lmfit.model.ModelResult):
        raise TypeError(
            "Input model_class must be a {0}, but it is {1}".format(
                lmfit.model.ModelResult.__name__, 
                model_result.__class__.__name__))
    if df.empty:
        raise ValueError("Input data frame df is empty")
        
    if fit_kws is None:
        fit_kws = {}
    if not 'param_fix' in fit_kws:
        fit_kws['param_fix'] = {pname for pname, param in model_result.params.items() if not param.vary}
    if not 'param_max' in fit_kws:
        fit_kws['param_max'] = {pname: param.max for pname, param in model_result.params.items()}
    if not 'param_max' in fit_kws:
        fit_kws['param_min'] = {pname: param.min for pname, param in model_result.params.items()}
    model_class = type(model_result.model)

    unique_units = pd.Series(df[unit].unique())
    grouped = df.groupby(unit)
    param_samples = [None] * nsamples

    for i in range(nsamples):
        sampled_units = unique_units.sample(frac=1, replace=True)
        _df = pd.concat(grouped.get_group(grp_id) for grp_id in sampled_units.values)
        assert (_df[unit].unique() == sampled_units.unique()).all()
        assert _df.shape == df.shape    
        model_fit = curveball.models.fit_model(_df, models=model_class, 
                                               PLOT=False, PRINT=False, 
                                               **fit_kws)[0]
        param_samples[i] = model_fit.best_values
    return pd.DataFrame(param_samples)


def sample_params(model_fit, nsamples, params=None, covar=None):
    """Random sample of parameter values from a truncated multivariate normal distribution defined by the 
    covariance matrix of the a model fitting result.

    Parameters
    ----------
    model_fit : lmfit.model.ModelResult
        the model fit that defines the sampled distribution
    nsamples : int
        number of samples to make
    params : dict, optional
        a dictionary of model parameter values; if given, overrides values from `model_fit`
    covar : numpy.ndarray, optional
        an array containing the parameters covariance matrix; if given, overrides values from `model_fit`

    Returns
    -------
    pandas.DataFrame
        data frame of samples; each row is a sample, each column is a parameter.

    See also
    --------
    bootstrap_params
    """
    if params is None:
        params = model_fit.params
    else:
        _params = model_fit.params.copy()
        for pname, pvalue in params.items():
            _params[pname].value = pvalue
        params = _params
    if covar is None:
        covar = model_fit.covar
    if covar is None:
        raise ValueError("Covariance matrix for {0} is invalid (None).".format(model_fit.model))
    if covar.ndim != 2:
        warn("Covariance matrix doesn't have 2 dimensions: \n{}".format(covar))
    w, h = covar.shape
    if w != h:
        warn("Covariance matrix is not square: \n{}".format(covar))
    names = [p.name for p in params.values() if p.vary]
    means = [p.value for p in params.values() if p.vary]
        
    param_samples = np.random.multivariate_normal(means, covar, nsamples)
    param_samples = pd.DataFrame(param_samples, columns=names)
    idx = np.zeros(nsamples) == 0
    for p in params.values():
        if not p.vary:
            param_samples[p.name] = p.value
        idx = idx & (param_samples[p.name] >= p.min) & (param_samples[p.name] <= p.max)
    if param_samples.shape[0] < nsamples:
        warn("Warning: truncated {0} parameter samples; please report at {1}, including the data and use case.".format(nsamples - param_samples.shape[0], "https://github.com/yoavram/curveball/issues"))
    param_samples = param_samples[idx]
    return param_samples
    


def noisify_normal_additive(data, std, rng=None):
    if not rng:
        rng = np.random
    return data +  rng.normal(0, std, data.shape)


def noisify_lognormal_multiplicative(data, std, random_seed=None):
    if random_seed is None:
        rng = np.random
    else:
        rng = np.random.RandomState(random_seed)
    return data * rng.lognormal(0, std, data.shape)


def randomize(t=12, y0=0.1, K=1.0, r=0.1, nu=1.0, q0=np.inf, v=np.inf, 
        func=curveball.baranyi_roberts_model.baranyi_roberts_function, reps=1, noise_std=0.02, 
        noise_func=noisify_lognormal_multiplicative, random_seed=None, as_df=True, 
        data_label='OD', time_label='Time', replicate_label='Well'):
    if isinstance(t, numbers.Number):
        t = np.linspace(0, t)
    y = func(t, y0, K, r, nu, q0, v)
    y.resize((len(t),))
    y = y.repeat(reps)
    y.resize((len(t), reps))
    well = np.arange(reps).repeat(len(t))#.resize(len(t), reps)
    if noise_std > 0:
        y = noise_func(y, noise_std, random_seed)
    y[y < 0] = 0.0
    y = y.flatten()
    t = t.repeat(reps)
    well = well.flatten()
    if as_df:
        return pd.DataFrame({data_label: y, time_label: t, replicate_label: well})
    else:
        return t, y


def lrtest(m0, m1, alfa=0.05):
    r"""Performs a likelihood ratio test on two nested models.

    For two models, one nested in the other 
    (meaning that the nested model estimated parameters are a subset of the nesting model), 
    the test statistic :math:`D` is:

    .. math::

        \Lambda = \Big( \Big(\frac{\sum{(X_i - \hat{X_i}(\theta_1))^2}}{\sum{(X_i - \hat{X_i}(\theta_0))^2}}\Big)^{n/2} \Big)

        D = -2 log \Lambda

        lim_{n \to \infty} D \sim \chi^2_{df=\Delta}


    where :math:`\Lambda` is the likelihood ratio, :math:`D` is the statistic, 
    :math:`X_{i}` are the data points, :math:`\hat{X_i}(\theta)` is the 
    model prediction with parameters :math:`\theta`, 
    :math:`\theta_i` is the parameters estimation for model :math:`i`, 
    :math:`n` is the number of data points, and :math:`\Delta` is the 
    difference in number of parameters between the models.

    The function compares between two :py:class:`lmfit.model.ModelResult` objects. 
    These are the results of fitting models to the same dataset 
    using the `lmfit <lmfit.github.io/lmfit-py>`_ package.

    The function compares between model fit `m0` and `m1` and assumes that `m0` is nested in `m1`, 
    meaning that the set of varying parameters of `m0` is a subset of the varying parameters of `m1`. 
    :py:attr:`lmfit.model.ModelResult.chisqr` is the sum of the square of the residuals of the fit. 
    :py:attr:`lmfit.model.ModelResult.ndata` is the number of data points. 
    :py:attr:`lmfit.model.ModelResult.nvarys` is the number of varying parameters.

    Parameters
    ----------
    m0, m1 : lmfit.model.ModelResult
        objects representing two model fitting results. `m0` is assumed to be nested in `m1`.
    alfa : float, optional
        test significance level, defaults to 0.05 = 5%.

    Returns
    -------
    prefer_m1 : bool
        should we prefer `m1` over `m0`
    pval : float
        the test p-value
    D : float
        the test statistic
    ddf : int 
        the number of degrees of freedom

    Notes
    ----------
    - `Generalised Likelihood Ratio Test Example <http://www.stat.sc.edu/~habing/courses/703/GLRTExample.pdf>`_.
    - `IPython notebook <http://nbviewer.ipython.org/github/yoavram/ipython-notebooks/blob/master/likelihood%20ratio%20test.ipynb>`_.
    """
    n0 = m0.ndata
    k0 = m0.nvarys
    chisqr0 = m0.chisqr
    assert chisqr0 > 0, chisqr0
    n1 = m1.ndata
    assert n0 == n1
    k1 = m1.nvarys
    chisqr1 = m1.chisqr
    assert chisqr1 > 0, chisqr1
    D = -n0 * (np.log(m1.chisqr) - np.log(m0.chisqr))
    assert D > 0, D
    ddf = k1 - k0
    assert ddf > 0, ddf
    pval = chisqprob(D, ddf)
    prefer_m1 = pval < alfa
    return prefer_m1, pval, D, ddf


def find_max_growth(model_fit, params=None, after_lag=True):
    r"""Estimates the maximum population and specific growth rates from the model fit.

    The function calculates the maximum population growth rate :math:`a=\max{\frac{dy}{dt}}` 
    as the derivative of the model curve and calculates its maximum. 
    It also calculates the maximum of the per capita growth rate :math:`\mu = \max{\frac{dy}{y \cdot dt}}`.
    The latter is more useful as a metric to compare different strains or treatments 
    as it does not depend on the population size/density.    

    For example, in the logistic model the population growth rate is a quadratic function of :math:`y` 
    so the maximum is realized when the 2nd derivative is zero:

    .. math::

        \frac{dy}{dt} = r y (1 - \frac{y}{K}) \Rightarrow

        \frac{d^2 y}{dt^2} = r - 2\frac{r}{K}y \Rightarrow

        \frac{d^2 y}{dt^2} = 0 \Rightarrow y = \frac{K}{2} \Rightarrow

        \max{\frac{dy}{dt}} = \frac{r K}{4}

    In contrast, the per capita growth rate a linear function of :math:`y` 
    and so its maximum is realized when :math:`y=y_0`:

    .. math::

        \frac{dy}{y \cdot dt} = r (1 - \frac{y}{K}) \Rightarrow

        \max{\frac{dy}{y \cdot dt}} = \frac{dy}{y \cdot dt}(y=y_0) = r (1 - \frac{y_0}{K})     

    Parameters
    ----------
    model_fit : lmfit.model.ModelResult
        the result of a model fitting procedure
    params : lmfit.parameter.Parameters, optional
        if provided, these parameters will override `model_fit`'s parameters
    after_lag : bool
        if true, only explore the time after the lag phase. Otherwise start at time zero. Defaults to :const:`True`.

    Returns
    -------
    t1 : float
        the time when the maximum population growth rate is achieved in the units of the `model_fit` ``Time`` variable.
    y1 : float
        the population size or density (OD) for which the maximum population growth rate is achieved.
    a : float
        the maximum population growth rate.
    t2 : float
        the time when the maximum per capita growth rate is achieved in the units of the `model_fit` `Time` variable.
    y2 : float
        the population size or density (OD) for which the maximum per capita growth rate is achieved.
    mu : float
        the the maximum specific (per capita) growth rate.

    See also
    --------
    find_max_growth_ci
    """
    if params is None:
        params = model_fit.params

    y0 = params['y0'].value
    K  = params['K'].value

    t0 = find_lag(model_fit) if after_lag else 0
    t0 = max(t0, 0)
    t1 = model_fit.userkws['t'].max()
    t = np.linspace(t0, t1)     
    def f(t): 
        return model_fit.model.eval(t=t, params=params)
    y = f(t)
    dfdt = derivative(f, t)

    a = dfdt.max()
    i = dfdt.argmax()
    t1 = t[i]
    y1 = y[i]

    dfdt_y = dfdt / y
    mu = dfdt_y.max()
    i = dfdt_y.argmax()
    t2 = t[i]
    y2 = y[i]
    
    return t1, y1, a, t2, y2, mu


def find_max_growth_ci(model_fit, param_samples, after_lag=True, ci=0.95):
    """Estimates a confidence interval for the maximum population/specific growth rates from the model fit.
    
    The maximum population/specific growth rate for each parameter sample is calculated.
    The confidence interval of the rate is the lower and higher percentiles such that 
    `ci` percent of the random rates are within the confidence interval.

    Parameters
    ----------
    model_fit : lmfit.model.ModelResult
        the result of a model fitting procedure
    param_samples : pandas.DataFrame
        parameter samples, generated using :function:`sample_params` or :function:`bootstrap_params`
    after_lag : bool
        if true, only explore the time after the lag phase. Otherwise start at time zero. Defaults to :const:`True`        
    ci : float, optional
        the fraction of lag durations that should be within the calculated limits. 0 < `ci` <, defaults to 0.95
    
    Returns
    -------
    low_a, high_a : float
        the lower and the higher boundaries of the confidence interval of the the maximum population growth rate in the units of the `model_fit` ``OD``/``Time`` (usually OD/hours).
    low_mu, high_mu : float
        the lower and the higher boundaries of the confidence interval of the the maximum specific growth rate in the units of the `model_fit` 1/``Time`` variable (usually 1/hours).


    See also
    --------
    find_max_growth
    """
    if not 0 <= ci <= 1:
        raise ValueError("ci must be between 0 and 1")
    nsamples = param_samples.shape[0]
    aa = np.zeros(nsamples)
    mumu = np.zeros(nsamples)        
    for i in range(param_samples.shape[0]):
        sample = param_samples.iloc[i,:]
        params = model_fit.params.copy()
        for k,v in params.items():
            if v.vary:
                params[k].set(value=sample[k])
        _, _, a, _, _, mu = find_max_growth(model_fit, params=params, after_lag=after_lag)
        aa[i] = a
        mumu[i] = mu

    margin = (1.0 - ci) * 50.0

    idx = np.isfinite(aa) & (aa >= 0)
    if not idx.all():
        warn("Warning: omitting {0} non-finite growth rate values".format(len(aa) - idx.sum()))
    aa = aa[idx]
    low_a = np.percentile(aa, margin)
    high_a = np.percentile(aa, ci * 100.0 + margin)
    assert high_a >= low_a, aa.tolist()

    idx = np.isfinite(mumu) & (mumu >= 0)
    if not idx.all():
        warn("Warning: omitting {0} non-finite growth rate values".format(len(mumu) - idx.sum()))
    mumu = mumu[idx]
    low_mu = np.percentile(mumu, margin)
    high_mu = np.percentile(mumu, ci * 100.0 + margin)
    assert high_mu >= low_mu, mumu.tolist()
    return low_a, high_a, low_mu, high_mu


def find_min_doubling_time(model_fit, params=None, PLOT=False):
    """Estimates the minimal doubling time from the model fit.

    The function evaluates a growth curves based on the model fit and supplied parameters (if any) 
    and calculates that time required to double the population density at each time point. 
    It then returns the minimal doubling time.

    Parameters
    ----------
    model_fit : lmfit.model.ModelResult
        the result of a model fitting procedure
    params : lmfit.parameter.Parameters, optional
        if provided, these parameters will override `model_fit`'s parameters
    PLOT : bool, optional
        if :const:`True`, the function will plot the Cook's distances of the wells and the threshold. # FIXME

    Returns
    -------
    min_dbl : float
        the minimal time it takes the population density to double.
    fig : matplotlib.figure.Figure
        if the argument `PLOT` was :const:`True`, the generated figure.
    ax : matplotlib.axes.Axes
        if the argument `PLOT` was :const:`True`, the generated axis.
    """
    if params is None:
        params = model_fit.params
    def f(t): 
        return model_fit.model.eval(t=t, params=params)

    t1 = model_fit.userkws['t'].max()
    t = np.linspace(0, t1, 1000)
    y = f(t)

    def find_point(point):
        return abs(y - point).argmin()

    imax = find_point(y.max() / 2) + 1
    doubling_times = np.zeros(imax)
    for i0, y0 in enumerate(y[:imax]):
        i1 = find_point(2 * y0)
        doubling_times[i0] = t[i1] - t[i0]
    min_dbl = doubling_times.min()

    if PLOT:
        fig, ax = plt.subplots(1, 1)
        ax.plot(t[:imax], doubling_times, '-')
        ax.axhline(min_dbl, ls='--', color='k')
        ax.set(xlabel='Time', ylabel='Doubling time')
        sns.despine()
        return min_dbl, fig, ax

    return min_dbl


def find_min_doubling_time_ci(model_fit, param_samples, ci=0.95):
    """Estimates a confidence interval for the minimal doubling time from the model fit.

    The minimal doubling time for each parameter sample is calculated.
    The confidence interval of the doubling time is the lower and higher percentiles such that 
    `ci` percent of the random lag durations are within the confidence interval.

    Parameters
    ----------
    model_fit : lmfit.model.ModelResult
        the result of a model fitting procedure
    param_samples : pandas.DataFrame
        parameter samples, generated using :function:`sample_params` or :function:`bootstrap_params`    
    ci : float, optional
        the fraction of doubling times that should be within the calculated limits. 0 < `ci` <, defaults to 0.95.
    
    Returns
    -------
    low, high : float
        the lower and the higher boundaries of the confidence interval of the minimal doubling times in the units of the `model_fit` ``Time`` variable (usually hours).

    See also
    --------
    find_min_doubling_time
    """
    if not 0 <= ci <= 1:
        raise ValueError("ci must be between 0 and 1")
    nsamples = param_samples.shape[0]
    dbls = np.zeros(nsamples)    
    for i in range(param_samples.shape[0]):
        sample = param_samples.iloc[i,:]
        params = model_fit.params.copy()
        for k, v in params.items():
            if v.vary:
                v.value = sample[k]
        dbls[i] = find_min_doubling_time(model_fit, params=params)

    margin = (1.0 - ci) * 50.0
    idx = np.isfinite(dbls) & (dbls >= 0)
    if not idx.all():
        warn("Warning: omitting {0} non-finite doubling time values".format(len(dbls) - idx.sum()))
    dbls = dbls[idx]
    low = np.percentile(dbls, margin)
    high = np.percentile(dbls, ci * 100.0 + margin)
    assert high >= low, dbls.tolist()
    return low, high


def find_K_ci(param_samples, ci=0.95):
    """Estimates a confidence interval for ``K``, the maximum population density from the model fit.

    The confidence interval of the doubling time is the lower and higher percentiles such that 
    `ci` percent of the max densities are within the confidence interval.

    Parameters
    ----------
    param_samples : pandas.DataFrame
        parameter samples, generated using :function:`sample_params` or :function:`bootstrap_params`    
    ci : float, optional
        the fraction of doubling times that should be within the calculated limits. 0 < `ci` <, defaults to 0.95.
    
    Returns
    -------
    low, high : float
        the lower and the higher boundaries of the confidence interval of the maximum population density.
    
    """
    if not 0 <= ci <= 1:
        raise ValueError("ci must be between 0 and 1")    
    Ks = param_samples['K']    
    
    margin = (1.0 - ci) * 50.0
    idx = np.isfinite(Ks) & (Ks >= 0)
    if not idx.all():
        warn("Warning: omitting {0} non-finite K values".format(len(Ks) - idx.sum()))
    Ks = Ks[idx]
    low = np.percentile(Ks, margin)
    high = np.percentile(Ks, ci * 100.0 + margin)
    assert high >= low, Ks.tolist()
    return low, high


def find_lag(model_fit, params=None):
    """Estimates the lag duration from the model fit.

    The function calculates the tangent line to the model curve at the point of maximum derivative (the inflection point). 
    The time when this line intersects with :math:`N_0` (the initial population size) 
    is labeled :math:`\lambda` and is called the lag duration time [fig2.2]_.

    Parameters
    ----------
    model_fit : lmfit.model.ModelResult
        the result of a model fitting procedure
    params : lmfit.parameter.Parameters, optional
        if provided, these parameters will override `model_fit`'s parameters

    Returns
    -------
    lam : float
        the lag phase duration in the units of the `model_fit` ``Time`` variable (usually hours).    

    References
    ----------
    .. [fig2.2] Fig. 2.2 pg. 19 in Baranyi, J., 2010. `Modelling and parameter estimation of bacterial growth with distributed lag time. <http://www2.sci.u-szeged.hu/fokozatok/PDF/Baranyi_Jozsef/Disszertacio.pdf>`_.

    See also
    --------
    find_lag_ci
    has_lag
    """
    if params is None:
        params = model_fit.params
    
    y0 = params['y0'].value
    K  = params['K'].value

    t = model_fit.userkws['t']
    t = np.linspace(t.min(), t.max())
    def f(t): 
        return model_fit.model.eval(t=t, params=params)
    y = f(t)    
    dfdt = derivative(f, t)
    idx = y > K / np.e
    if idx.sum() == 0:
        warn("All values are below K/e")
        return np.nan
    t = t[idx]
    y = y[idx]
    dfdt = dfdt[idx]

    a = dfdt.max()
    i = dfdt.argmax()
    t1 = t[i]
    y1 = y[i]
    b = y1 - a * t1
    lam = (y0 - b) / a
    return lam


def find_lag_ci(model_fit, param_samples, ci=0.95):
    """Estimates a confidence interval for the lag duration from the model fit.

    The lag duration for each parameter sample is calculated.
    The confidence interval of the lag is the lower and higher percentiles such that 
    `ci` percent of the random lag durations are within the confidence interval.

    Parameters
    ----------
    model_fit : lmfit.model.ModelResult
        the result of a model fitting procedure
    param_samples : pandas.DataFrame
        parameter samples, generated using :function:`sample_params` or :function:`bootstrap_params`    
    ci : float, optional
        the fraction of lag durations that should be within the calculated limits. 0 < `ci` <, defaults to 0.95.
    
    Returns
    -------
    low, high : float
        the lower and the higher boundaries of the confidence interval of the lag phase duration in the units of the `model_fit` ``Time`` variable (usually hours).

    See also
    --------
    find_lag
    has_lag    
    """    
    if not 0 <= ci <= 1:
        raise ValueError("ci must be between 0 and 1")
    nsamples = param_samples.shape[0]
    lags = np.zeros(nsamples)    
    for i in range(param_samples.shape[0]):
        sample = param_samples.iloc[i,:]
        params = model_fit.params.copy()
        for k,v in params.items():
            if v.vary:
                params[k].set(value=sample[k])
        lags[i] = find_lag(model_fit, params=params)

    margin = (1.0 - ci) * 50.0
    idx = np.isfinite(lags)
    if not idx.all():
        warn("Warning: omitting {0} non-finite lag values".format(len(lags) - idx.sum()))
    lags = lags[idx]
    idx = (lags >= 0)
    if not idx.all():
        warn("Warning: omitting {0} negative lag values".format(len(lags) - idx.sum()))
    if not idx.any(): # no legal lag values left
        return np.nan, np.nan, np.nan
    lags = lags[idx]
    low = np.percentile(lags, margin)
    high = np.percentile(lags, ci * 100.0 + margin)
    assert high >= low, lags.tolist()
    return low, high


def has_lag(model_fits, alfa=0.05, PRINT=False):
    r"""Checks if if the best fit has statisticaly significant lag phase :math:`\lambda > 0`.

    If the best fitted model doesn't has a lag phase to begin with, return :const:`False`. 
    This includes the logistic model and Richards model.

    Otherwise, a likelihood ratio test will be perfomed with nesting determined according to Figure 1. 
    The null hypothesis of the test is that :math:`\frac{1}{v} = 0` , 
    i.e. the adjustment rate :math:`v` is infinite and therefore there is no lag phase.

    The function will return :const:`True` if the null hypothesis is rejected, 
    otherwise it will return :const:`False`.

    Parameters
    ----------
    model_fits : sequence of lmfit.model.ModelResult
        the results of several model fitting procedures, ordered by their statistical preference. Generated by :py:func:`fit_model`.
    alfa : float, optional
        test significance level, defaults to 0.05 = 5%.
    PRINT : bool, optional
        if :const:`True`, the function will print the result of the underlying statistical test; defaults to :const:`False`.

    Returns
    -------
    bool
        the result of the hypothesis test. :const:`True` if the null hypothesis was rejected and the data suggest that there is a significant lag phase.

    Raises
    ------
    ValueError
        raised if the fittest of the :py:class:`lmfit.model.ModelResult` objects in `model_fits` is of an unknown model.
    """
    m1 = model_fits[0]
    if np.isposinf(m1.best_values.get('q0', np.inf)) and np.isposinf(m1.best_values.get('v', np.inf)):
        if PRINT:
            print('H1 model has no lag')
        return False
        
    try:
        m0_model_class = m1.model.nested_models['lag']
    except KeyError:
        raise ValueError("The best fit model {0} has no nested model for testing lag".format(m1.model.name))
    try:
        m0 = [m for m in model_fits if isinstance(m.model, m0_model_class)][0]
    except IndexError:
        raise ValueError("No {0} in model results.".format(m0_model_class.name))
    
    prefer_m1, pval, D, ddf = lrtest(m0, m1, alfa=alfa)
    if PRINT:
        print("Tested H0: %s vs. H1: %s; D=%.2g, ddf=%d, p-value=%.2g" % (m0.model.name, m1.model.name, D, ddf, pval))    
    return prefer_m1


def has_nu(model_fits, alfa=0.05, PRINT=False):
    r"""Checks if if the best fit has :math:`\nu \ne 1` and if so if that is statisticaly significant.

    If the best fitted model has :math:`\nu = 1` to begin with, return :const:`False`. This includes the logistic model.
    Otherwise, a likelihood ratio test will be perfomed with nesting determined according to Figure 1. 
    The null hypothesis of the test is that :math:`\nu = 1`; if it is rejected than the function will return :const:`True`.
    Otherwise it will return :const:`False`.

    Parameters
    ----------
    model_fits : list lmfit.model.ModelResult
        the results of several model fitting procedures, ordered by their statistical preference. Generated by :py:func:`fit_model`.
    alfa : float, optional
        test significance level, defaults to 0.05 = 5%.
    PRINT : bool, optional
        if :const:`True`, the function will print the result of the underlying statistical test; defaults to :const:`False`.

    Returns
    -------
    bool
        the result of the hypothesis test. :const:`True` if the null hypothesis was rejected and the data suggest that :math:`\nu` is significantly different from one.

    Raises
    ------
    ValueError
        raised if the fittest of the :py:class:`lmfit.model.ModelResult` objects in `model_fits` is of an unknown model.
    """
    m1 = model_fits[0]
    if m1.best_values.get('nu', 1.0) == 1.0:
        return False

    try:
        m0_model_class = m1.model.nested_models['nu']
    except KeyError:
        raise ValueError("The best fit model {} has no nested model for testing nu".format(m1.model.name))
    try:
        m0 = [m for m in model_fits if isinstance(m.model, m0_model_class)][0]
    except IndexError:
        raise ValueError("No {} in model results.".format(m0_model_class.name))

    prefer_m1, pval, D, ddf = lrtest(m0, m1, alfa=alfa)
    if PRINT:
        msg = "Tested H0: %s (nu=%.2g) vs. H1: %s (nu=%.2g); D=%.2g, ddf=%d, p-value=%.2g"  
        print(msg % (m0.model.name, m0.best_values.get('nu', 1), m1.model.name, m1.best_values.get('nu', 1), D, ddf, pval))
    return prefer_m1


def make_Dfun(model, params):
    expr, t, args = model.get_sympy_expr(params)
    partial_derivs = [None]*len(args)
    for i,x in enumerate(args):
        dydx = expr.diff(x)
        dydx = sympy.lambdify(args=(t,) + args, expr=dydx, modules="numpy")
        partial_derivs[i] = dydx        
    def Dfun(params, y, a, t):
        values = [ par.value for par in params.values() if par.vary ]
        res = np.array([dydx(t, *values) for dydx in partial_derivs])
        expected_shape = (len(values), len(t))
        if res.shape != expected_shape:
            raise TypeError("Dfun result shape for {0} is incorrect, expected {1} but it is {2}.".format(model.name, expected_shape, res.shape))
        return res
    return Dfun


def cooks_distance(df, model_fit, use_weights=True):
    """Calculates Cook's distance of each well given a specific model fit. 

    Cook's distance is an estimate of the influence of a data curve when performing model fitting; 
    it is used to find wells (growth curve replicates) that are suspicious as outliers.
    The higher the distance, the more suspicious the curve.

    Parameters
    ----------
    df : pandas.DataFrame
        growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
    model_fit : lmfit.model.ModelResult
        result of model fitting procedure
    use_weights : bool, optional
        should the function use standard deviation across replicates as weights for the fitting procedure, defaults to :const:`True`.

    Returns
    -------
    dict
        a dictionary of Cook's distances: keys are wells (from the `Well` column in `df`), values are Cook's distances.

    Notes
    -----
    `Wikipedia <https://en.wikipedia.org/wiki/Cook's_distance>`_
    """
    p = model_fit.nvarys
    MSE = model_fit.chisqr / model_fit.ndata
    wells = df.Well.unique()
    D = {}
    
    for well in wells:    
        _df = df[df.Well != well]
        time = _df['Time'].values
        OD = _df['OD'].values
        weights =  calc_weights(_df) if use_weights else None
        model_fit_i = copy.deepcopy(model_fit)
        model_fit_i.fit(data=OD, t=time, weights=weights)
        D[well] = model_fit_i.chisqr / (p * MSE)
    return D


def find_outliers(df, model_fit, deviations=2, use_weights=True, ax=None, PLOT=False):
    """Find outlier wells in growth curve data.

    Uses the Cook's distance approach (`cooks_distance`); 
    values of Cook's distance that are `deviations` standard deviations **above** the mean
    are defined as outliers.

    Parameters
    ----------
    df : pandas.DataFrame
        growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
    model_fit : lmfit.model.ModelResult
        result of model fitting procedure
    deviations : float, optional
        the number of standard deviations that defines an outlier, defaults to 2.
    use_weights : bool, optional
        should the function use standard deviation across replicates as weights for the fitting procedure, defaults to :const:`True`.
    ax : matplotlib.axes.Axes, optional
        an axes to plot into; if not provided, a new one is created.
    PLOT : bool, optional
        if :const:`True`, the function will plot the Cook's distances of the wells and the threshold.

    Returns
    -------
    outliers : list
        the labels of the outlier wells.
    fig : matplotlib.figure.Figure
        if the argument `PLOT` was :const:`True`, the generated figure.
    ax : matplotlib.axes.Axes
        if the argument `PLOT` was :const:`True`, the generated axis.
    """
    D = cooks_distance(df, model_fit, use_weights=use_weights)
    D = sorted(D.items()) # TODO SortedDict?
    distances = [x[1] for x in D]        
    dist_mean, dist_std = np.mean(distances), np.std(distances)
    outliers = [well for well,dist in D if dist > dist_mean + deviations * dist_std]
    if PLOT:
        if ax is None:
            fig,ax = plt.subplots(1,1)
        else:
            fig = ax.get_figure()
        wells = [x[0] for x in D]            
        ax.stem(distances, linefmt='k-', basefmt='')
        ax.axhline(y=dist_mean, ls='-', color='k')
        ax.axhline(y=dist_mean + deviations * dist_std, ls='--', color='k')
        ax.axhline(y=dist_mean - deviations * dist_std, ls='--', color='k')
        ax.set_xticks(range(len(wells)))
        ax.set_xlim(-0.5, len(wells) - 0.5)
        ax.set_xticklabels(wells, rotation=90)
        ax.set_xlabel('Well')
        ax.set_ylabel("Cook's distance")
        sns.despine()
        return outliers,fig,ax
    return outliers


def find_all_outliers(df, model_fit, deviations=2, max_outlier_fraction=0.1, use_weights=True, PLOT=False):    
    """Iteratively find outlier wells in growth curve data.

    At each iteration, calls :py:func:`find_outliers`.
    Iterations stop when no more outliers are found or when the fraction of wells defined as outliers
    is above `max_outlier_fraction`.
        
    Parameters
    ----------
    df : pandas.DataFrame
        growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
    model_fit : lmfit.model.ModelResult
        result of model fitting procedure
    deviations : float, optional
        the number of standard deviations that defines an outlier, defaults to 2.
    max_outlier_fraction : float, optional
        maximum fraction of wells to define as outliers, defaults to 0.1 = 10%.
    use_weights : bool, optional
        should the function use standard deviation across replicates as weights for the fitting procedure, defaults to :const:`True`.
    ax : matplotlib.axes.Axes
        an axes to plot into; if not provided, a new one is created.
    PLOT : bool, optional
        if :const:`True`, the function will plot the Cook's distances of the wells and the threshold.

    Returns
    -------
    outliers : list
        a list of lists: the list nested at index *i* contains the labels of the outlier wells found at iteration *i*.
    fig : matplotlib.figure.Figure
        if the argument `PLOT` was :const:`True`, the generated figure.
    ax : matplotlib.axes.Axes
        if the argument `PLOT` was :const:`True`, the generated axis.
    """
    outliers = []
    num_wells = len(df.Well.unique())
    df = copy.deepcopy(df)
    if PLOT:
        fig = plt.figure()
        o, fig, ax = find_outliers(df, model_fit, deviations=deviations, use_weights=use_weights, ax=fig.add_subplot(), PLOT=PLOT)
    else:
        o = find_outliers(df, model_fit, deviations=deviations, use_weights=use_weights, PLOT=PLOT)
    outliers.append(o)
    while len(outliers[-1]) != 0 and len(sum(outliers, [])) <  max_outlier_fraction * num_wells:
        df = df[~df.Well.isin(outliers[-1])]
        assert df.shape[0] > 0, df.shape[0] 
        model_fit = fit_model(df, use_weights=use_weights, PLOT=False, PRINT=False)[0]
        if PLOT:
            o, fig, ax = find_outliers(df, model_fit, deviations=deviations, use_weights=use_weights, ax=fig.add_subplot(), PLOT=PLOT)            
        else:
            o = find_outliers(df, model_fit, deviations=deviations, use_weights=use_weights, PLOT=PLOT)
        outliers.append(o)
    if PLOT:
        return outliers[:-1],fig,ax
    return outliers[:-1]


def calc_weights(df, PLOT=False):
    """Calculate weights for the fittiing procedure based on the standard deviations at each time point.

    If there is more than one replicate, use the standard deviations as weight.
    Warn about NaN and infinite values.

    Parameters
    ----------
    df : pandas.DataFrame
        data frame with `Time` and `OD` columns
    PLOT : bool, optional
        if :const:`True`, plot the weights by time

    Returns
    -------
    weights : np.ndarray
        array of weights, calculated as the inverse of the standard deviations at each time point
    fig : matplotlib.figure.Figure
        if the argument `PLOT` was :const:`True`, the generated figure.
    ax : matplotlib.axes.Axes
        if the argument `PLOT` was :const:`True`, the generated axis.
    """
    deviations = df.groupby('Time')['OD'].transform(lambda x: np.repeat(x.std(), len(x))).values
    if np.isnan(deviations).any():
        warn("NaN in deviations, can't use weights")
        weights = None
    else:
        weights = 1.0 / deviations
        # if any weight is nan, raise error
        idx = np.isnan(weights)
        if idx.any():
            raise ValueError("NaN weights are illegal, indices: {0}".format(idx))
        # if any weight is infinite, change to the max
        idx = np.isinf(weights)
        if idx.all():
            warn("All weights are infinite, proceeding without weights)")
            weights = None
        elif idx.any():
            warn("Found infinite weight, changing to maximum ({0} occurences)".format(idx.sum()))
            weights[idx] = weights[~idx].max()
    if PLOT:
        fig, ax = plt.subplots(1, 1)
        ax.plot(df.Time, weights, 'o')
        ax.set_xlabel('Time')
        ax.set_ylabel('Weight')
        sns.despine()
        return weights, fig, ax
    return weights


def information_criteria_weights(results):
    r"""Calculate weighted AIC and BIC for model results.

    .. :math::

        w_i = \frac{e^{-\frac{x_i - \bar{x}}{2}}}{\sum_j{e^{-\frac{x_j - \bar{x}}{2}}}}

    where :math:`w_i` is the weighted measure for model result :math:`i` and
    :math:`x_i` is the AIC or BIC of model result :math:`i`.

    Parameters
    ----------
    results : sequence of lmfit.model.ModelResult
        use the :py:attr:`lmfit.attr.ModelResult.aic` and :py:attr:`lmfit.model.ModelResult.bic` attributes to add a `weighted_aic` and `weighted_bic` attribute.

    Notes
    -----
    - `Weighted AIC explained <http://stats.stackexchange.com/questions/35010/aic-bic-number-interpretation/44360#44360>`_.
    """
    aics = np.array([m.aic for m in results])
    bics = np.array([m.bic for m in results])
    weighted_aics = np.exp(-0.5 * (aics - aics.min()))
    weighted_aics /= weighted_aics.sum()
    weighted_bics = np.exp(-0.5 * (bics - bics.min()))
    weighted_bics /= weighted_bics.sum()
    for m, b, a in zip(results, weighted_bics, weighted_aics):
        m.weighted_aic = a
        m.weighted_bic = b


def nvarys(params):
    return len([p for p in params.values() if p.vary])


def fit_exponential_growth_phase(t, N, k=2):
    r"""Fits an exponential model to the exponential growth phase.

    Fits a polynomial p(t)~N, finds tmax the time of the maximum of the derivative dp/dt,
    and fits a linear function to log(N) around tmax.
    The resulting slope (a) and intercept (b) are the parameters of the exponential model:

    .. math::

        N(t) = N_0 e^{at}

        N_0 = e^b

    Arguments
    ---------
    t : np.ndarray
        time
    N : np.ndarray
        `N[i]` is the population size at time `t[i]`
    k : int
        number of points to take around tmax, defaults to 2 for a total of 5 points

    Returns
    -------
    slope : float
        slope of the linear regression, a
    intercept : float
        intercept of the linear regression, b
    """
    N_smooth = smooth(t, N)
    dNdt = derivative(N_smooth, t)
    imax = dNdt.argmax()
    idx = np.arange(imax - k, imax + k)
    slope, intercept = linregress(t[idx], np.log(N[idx]))[:2]
    return slope, intercept


def fit_model(df, param_guess=None, param_min=None, param_max=None, param_fix=None, 
              models=None, use_weights=False, use_Dfun=False, method='leastsq', ax=None, PLOT=True, PRINT=True):
    r"""Fit and select a growth model to growth curve data.

    This function fits several growth models to growth curve data (``OD`` as a function of ``Time``).

    Parameters
    ----------
    df : pandas.DataFrame
        growth curve data, see :py:mod:`curveball.ioutils` for a detailed definition.
    param_guess : dict, optional
        a dictionary of parameter guesses to use (key: :py:class:`str` of param name; value: :py:class:`float` of param guess).
    param_min : dict, optional
        a dictionary of parameter minimum bounds to use (key: :py:class:`str` of param name; value: :py:class:`float` of param min bound).
    param_max : dict, optional
        a dictionary of parameter maximum bounds to use (key: :py:class:`str` of param name; value: :py:class:`float` of param max bound).
    param_fix : set, optional
        a set of names (:py:class:`str`) of parameters to fix rather then vary, while fitting the models.
    use_weights : bool, optional
        should the function use the deviation across replicates as weights for the fitting procedure, defaults to :const:`False`.
    use_Dfun : bool, optional
        should the function calculate the partial derivatives of the model functions to be used in the fitting procedure, defaults to :const:`False`.
    models : one or more model classes, optional
        model classes (not instances) to use for fitting; defaults to all model classes in `curveball.baranyi_roberts_model`.
    method : str, optional
        the minimization method to use, defaults to `leastsq`, 
        can be anything accepted by :py:func:`lmfit.minimizer.Minimizer.minimize` or :py:func:`lmfit.minimizer.Minimizer.scalar_minimize`.
    ax : matplotlib.axes.Axes, optional
        an axes to plot into; if not provided, a new one is created.
    PLOT : bool, optional
        if :const:`True`, the function will plot the all model fitting results.
    PRINT : bool, optional
        if :const:`True`, the function will print the all model fitting results.

    Returns
    -------
    models : list of lmfit.model.ModelResult
        all model fitting results, sorted by increasing BIC (a measure of fit quality).
    fig : matplotlib.figure.Figure
        figure object.
    ax : numpy.ndarray
        array of :py:class:`matplotlib.axes.Axes` objects, one for each model result, with the same order as `models`.

    Raises
    ------
    TypeError
        if one of the input parameters is of the wrong type (not guaranteed).
    ValueError
        if the input is bad, for example, `df` is empty (not guaranteed).
    AssertionError
        if any of the intermediate calculated values are inconsistent (for example, ``y0<0``).

    Examples
    --------
    >>> import curveball
    >>> import pandas as pd
    >>> import matplotlib.pyplot as plt
    >>> plate = pd.read_csv('plate_templates/G-RG-R.csv')
    >>> df = curveball.ioutils.read_tecan_xlsx('data/Tecan_280715.xlsx', label='OD', plate=plate)
    >>> green_models = curveball.models.fit_model(df[df.Strain == 'G'], PLOT=True)
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("Input df must be a %s, but it is %s" % (pd.DataFrame.__name__, df.__class__.__name__))
    if df.shape[0] == 0:
        raise ValueError("No rows in input df")
    
    if param_guess is None:
        param_guess = dict()
    if param_fix is None:
        param_fix = set()

    df = df.sort_values(by=['Time', 'OD'])
    time = df['Time'].values
    OD = df['OD'].values
    weights =  calc_weights(df) if use_weights else None
    # TODO why should we use weights if we use the whole data set?
    ODerr = df.groupby('Time').OD.transform(lambda x: np.repeat(x.std(), len(x))).values
   
    if models is None:
        models = get_models(curveball.baranyi_roberts_model)
    elif is_model(models):
        models = [models]
    results = [None] * len(models)
    for i, model_class in enumerate(models):
        model = model_class()
        params = model.guess(data=OD, t=time, param_guess=param_guess, param_min=param_min, param_max=param_max, param_fix=param_fix)    
        fit_kws = {'Dfun': make_Dfun(model, params), "col_deriv":True} if use_Dfun else {}        
        model_result = model.fit(data=OD, t=time, params=params, weights=weights, fit_kws=fit_kws, method=method)
        results[i] = model_result

    # sort by increasing BIC
    information_criteria_weights(results)
    results.sort(key=lambda m: m.bic)

    if PRINT:
        print(results[0].fit_report(show_correl=False))
    if PLOT:        
        dy = df['OD'].max() / 50.0
        dx = df['Time'].max() / 25.0
        columns = min(3, len(results))
        rows = int(np.ceil(len(results) / columns))
        w = max(8, 4 * columns)
        h = max(6, 3*rows)
        fig, ax = plt.subplots(rows, columns, sharex=True, sharey=True, figsize=(w, h))
        if not hasattr(ax, '__iter__'):
            ax = np.array(ax, ndmin=2)
        for i,fit in enumerate(results):
            row = i // columns
            col = i % columns
            _ax = ax[row, col]
            vals = fit.best_values
            fit.plot_fit(ax=_ax, datafmt='.', data_kws={'alpha':0.3}, fit_kws={'lw': 4}, yerr=ODerr)
            _ax.axhline(y=vals.get('y0', 0), color='k', ls='--')
            _ax.axhline(y=vals.get('K', 0), color='k', ls='--')          
            title = '%s %dp\nBIC: %.3f\ny0=%.2f, K=%.2f, r=%.2g\n' + r'$\nu$=%.2g, $q_0$=%.2g, v=%.2g'
            title = title % (fit.model.name, fit.nvarys, fit.bic, vals.get('y0', np.nan), vals.get('K', np.nan), vals.get('r', np.nan), vals.get('nu', np.nan), vals.get('q0', np.nan), vals.get('v', np.nan))
            _ax.set_title(title)
            _ax.get_legend().set_visible(False)
            _ax.set_xlabel('')
            _ax.set_ylabel('')
            if col == 0:
                _ax.set_ylabel('OD')
            if row == rows - 1:
                _ax.set_xlabel('Time')
        _ax.set_xlim(0, 1.1 * df['Time'].max())
        _ax.set_ylim(0.9 * df['OD'].min(), 1.1 * df['OD'].max())
        sns.despine()
        fig.tight_layout()
        return results, fig, ax
    return results


if __name__ == '__main__':    
    # simulate 30 growth curves
    df = randomize(t=12, y0=0.12, K=0.56, r=0.8, nu=1.8, q0=0.2, v=0.8, reps=30, noise_std=0.04, random_seed=0)

    # fit models to growth curves
    results, fig, ax = fit_model(df, use_Dfun=True, PLOT=True, PRINT=True)    
    plt.savefig('test_models.png')
    plt.show()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from builtins import map
import sys
import os.path
from importlib import resources
import glob
import warnings
# catch some future warnings, mostly caused by matplotlib
warnings.simplefilter(action="ignore", category=FutureWarning)
import curveball
import numpy as np
import pandas as pd
from scipy.integrate import odeint
import click
import zipfile
from openpyxl.utils.exceptions import InvalidFileException
import matplotlib.pyplot as plt
import seaborn as sns
sns.set_style("ticks")


VERBOSE = False
PLOT = True
PROMPT = True
ERROR_COLOR = 'red'
INFO_COLOR = 'white'
file_extension_handlers = {
	'.mat': curveball.ioutils.read_tecan_mat, 
	'.xlsx': curveball.ioutils.read_tecan_xlsx,
	'.csv': curveball.ioutils.read_curveball_csv,
}


def echo_error(message):
	click.secho("Error: %s" % message, fg=ERROR_COLOR)


def echo_info(message):
	if VERBOSE:
		click.secho(message, fg=INFO_COLOR)


def ioerror_to_click_exception(io_error):
	raise click.FileError(io_error.filename, hint=io_error.message)


def to_dict(ctx, param, value):
    return dict(value)


def to_set(ctx, param, value):
    return set(value)


def get_filename(filepath):
	"""Get a file name out of a file path.

	Parameters
	----------
	filepath : str

	Returns
	-------
	str
		filename
	"""
	if filepath is None:
		return ''
	filename = os.path.split(filepath)[-1]
	if filename is None:
		return ''
	return filename


def find_plate_file(plate_folder, plate_file):
	"""Finds a plate file, either in the current working dir or in the package data resources.

	Parameters
	----------
	plate_file : str
		the filename of the plate file, may include absolute or relative path

	Returns
	-------
	str
		the full path of the plate file.
	"""
	plate_path = os.path.join(plate_folder, plate_file)
	if not os.path.exists(plate_path):
		# if plate path doesn't exist try to get it from package data
		plate_path = _resource_path(plate_folder, plate_file)
	if not os.path.exists(plate_path):
		raise click.FileError(plate_path, hint="can't find file.")
	return plate_path


def load_plate(plate_path):
	"""Loads a plate template from a CSV file.

	Parameters
	----------
	plate_path : str
		full or relative path to the plate template file.

	Returns
	-------
	pandas.DataFrame
		the plate template in tidy data format (see :py:mod:`.ioutils`).

	See also
	--------
	find_plate_file
	"""	
	try:
		plate = pd.read_csv(plate_path)
	except IOError as e:
		ioerror_to_click_exception(e)
	except pd.parser.CParserError as e:
		raise click.FileError(plate_path, hint="parser error, probably not a CSV file, {0}".format(e.args[0]))
	return plate


def where(ctx, param, value):
	"""Prints the path where Curveball is installed and exits. 

	Parameters are generally ignored; 
	if `value` or `ctx.resilient_parsing` are not empty/:const:`False`/:const:`None`, 
	the function returns without doing anything.
	"""
	if not value or ctx.resilient_parsing:
		return
	path = curveball.__file__
	folder = os.path.split(path)[0]
	click.secho(click.format_filename(folder))
	ctx.exit()


@click.group()
@click.option('-v/-V', '--verbose/--no-verbose', default=False)
@click.option('-l/-L', '--plot/--no-plot', default=True)
@click.option('-p/-P', '--prompt/--no-prompt', default=False)
@click.option('--where', is_flag=True, default=False, is_eager=True, callback=where, help='prints the path where Curveball is installed')
@click.version_option(version=curveball.__version__, prog_name=curveball.__name__)
def cli(verbose, plot, prompt, where):
	"""Main entry point to curveball

	To get help for the parameters, run:

	>>> curveball --help
	"""
	global VERBOSE
	VERBOSE = verbose
	global PLOT
	PLOT = plot
	global PROMPT
	PROMPT = prompt
	if VERBOSE:
		click.secho('=' * 40, fg='cyan')
		click.secho('Curveball %s' % curveball.__version__, fg='cyan')
		click.secho('=' * 40, fg='cyan')


@click.option('--plate_folder', default='plate_templates', help='plate templates default folder', type=click.Path())
@click.option('--plate_file', default='checkerboard.csv', help='plate templates csv file')
@click.option('-o', '--output_file', default='-', help='output file path', type=click.File(mode='w', lazy=True))
@click.option('--list', is_flag=True, default=False, help='list plate templates in the default folder')
@click.option('--show', is_flag=True, default=False, help='display the plate template as an image')
@cli.command()
def plate(plate_folder, plate_file, output_file, list, show):
	"""Read and print a plate template from a plate template CSV file.

	To get help for the parameters, run:

	>>> curveball plate --help
	"""
	if list:
		files = [p.name for p in resources.files('plate_templates').iterdir()]
		files = [fn for fn in files if os.path.splitext(fn)[-1].lower() == '.csv']
		files = os.linesep.join(files)
		click.echo(files)
		return
	plate_path = find_plate_file(plate_folder, plate_file)
	plate = load_plate(plate_path)
	if show:
		fig, ax = curveball.plots.plot_plate(plate)
		if output_file.name == '-':
			plt.show()
		else:
			fig.savefig(output_file.name)
	else:
		plate.to_csv(output_file, index=False)
	if output_file.name != '-':
		echo_info("Wrote output to {0}".format(click.format_filename(output_file.name)))


@click.argument('filepath', type=click.Path(exists=True, readable=True))
@click.option('--analyse_file', default='-', help='analyse output csv file path', type=click.Path(exists=True, readable=True))
@click.option('--ref_strain', type=str, help='strain used as reference, w=1')
@click.option('--total', type=str, help='strain used as total of mixed growth')
@click.option('--max_time', default=np.inf, help='omit data after max_time hours')
@click.option('--plate_folder', default='plate_templates', help='plate templates default folder', type=click.Path())
@click.option('--plate_file', default='checkerboard.csv', help='plate templates csv file')
@click.option('-o', '--output_file', default='-', help='output csv file path', type=click.File(mode='w', lazy=True))
@cli.command()
def fit_compete(filepath, analyse_file, output_file, ref_strain, total, max_time, plate_folder, plate_file):
	"""Estimate competition coefficients using mixed growth curves data."""
	param_names = ['K', 'r', 'nu', 'q0', 'v']
	ode = curveball.competitions.baranyi_roberts_yr

	fn, _ = os.path.splitext(filepath)
	analyse_results = pd.read_csv(analyse_file)
	idx = analyse_results['strain'] == ref_strain
	params1 = analyse_results[idx]
	idx = ~(idx | (analyse_results['strain'] == total))
	params2 = analyse_results[idx]
	y0 = (params1['y0'].values[0], params2['y0'].values[0])
	params1 = {v: params1[v].values[0] for v in param_names}
	params2 = {v: params2[v].values[0] for v in param_names}
	params = [(params1[v], params2[v]) for v in param_names]

	plate_path = find_plate_file(plate_folder, plate_file)
	plate = load_plate(plate_path)
	df = _load_file(filepath, max_time, plate)
	total_df = df[df.Strain == total]
	total_t = np.unique(total_df.Time)
	total_mean = total_df.groupby('Time').OD.mean().values
	total_y0 = total_df.loc[total_df.Time == 0, 'OD'].mean()
	y0sum = y0[0] + y0[1]
	y0 = (total_y0 * y0[0] / y0sum, total_y0 * y0[1] / y0sum)

	retval = curveball.competitions.fit_and_compete(
		params1,
		params2,
		total_df,
		ode=ode,
		y0=y0,
		aguess=(1, 1),
		PLOT=PLOT,
		fixed=True,
		method='bfgs'
	)
	if PLOT:
		t, y, a, fig, ax = retval
		plot_fn = fn + '_fit-compete.png'
		fig.savefig(plot_fn)
		echo_info("Wrote plot to %s" % click.format_filename(plot_fn))
	else:
		t, y, a = retval
	params.append(a)

	MSE = ((odeint(ode, y0, total_t, args=tuple(params)).sum(axis=1) - total_mean)**2).mean()

	result = dict()
	result['a1'] = a[0]
	result['a2'] = a[1]
	result['MSE'] = MSE

	output_table = pd.DataFrame([result]).transpose()
	output_table.to_csv(output_file)
	if VERBOSE and output_file.name != '-':
		click.secho("Wrote output to %s" % output_file.name, fg='green')


@click.argument('path', type=click.Path(exists=True, readable=True))
@click.option('--plate_folder', default='plate_templates', help='plate templates default folder', type=click.Path())
@click.option('--plate_file', default='checkerboard.csv', help='plate templates csv file')
@click.option('-o', '--output_file', default='-', help='output csv file path', type=click.File(mode='w', lazy=True))
@click.option('--blank_strain', default='0', type=str, help='blank strain for background calibration')
@click.option('--ref_strain', default='1',  type=str, help='reference strain for competitions')
@click.option('--max_time', default=np.inf, help='omit data after max_time hours')
@click.option('--guess', type=(str, float), multiple=True, callback=to_dict, help='set the initial guess for a parameter')
@click.option('--param_min', type=(str, float), multiple=True, callback=to_dict, help='set the minimum allowed value for a parameter')
@click.option('--param_max', type=(str, float), multiple=True, callback=to_dict, help='set the maximum allowed value for a parameter')
@click.option('--param_fix', type=str, multiple=True, callback=to_set, help='fix a parameter to it\'s initial guess')
@click.option('--weights/--no-weights', default=False, help="use weights for the fitting procedure")
@click.option('--ci/--no-ci', default=False, help="find confidence intervals for lag and max growth rate")
@click.option('--nsamples', default=1000, help="number of bootstrap samples to use, only applicable when using --ci")
@cli.command()
def analyse(path, output_file, plate_folder, plate_file, blank_strain, ref_strain, max_time, guess, param_min, param_max, param_fix, weights, ci, nsamples):
	"""Analyse growth curves data using Curveball.

	To get help for the parameters, run:

	>>> curveball plate --help
	"""
	results = []
	plate_path = find_plate_file(plate_folder, plate_file)

	if VERBOSE:
		click.echo('- Processing %s' % click.format_filename(path))		
		click.echo('- Using plate template from %s' % click.format_filename(plate_path))
		click.echo('- Blank strain: %s; Reference strain: %s' % (blank_strain, ref_strain))
		click.echo('- Omitting data after %.2f hours' % max_time)
		click.echo('-' * 40)
	
	plate = load_plate(plate_path)
	plate.Strain = list(map(str, plate.Strain))
	plate_strains = plate.Strain.unique().tolist()	
	if PROMPT:
		fig,ax = curveball.plots.plot_plate(plate)
		fig.show()
		click.echo("Plate with %d strains: %s" % (len(plate_strains), ', '.join(plate_strains)))
		click.confirm('Is this the plate you wanted?', default=False, abort=True, show_default=True)
	if os.path.isdir(path):
		files = glob.glob(os.path.join(path, '*'))
		#files = [os.path.join(path, fn) for fn in files]
	else:
		files = glob.glob(path)
	
	files = [fn for fn in files if os.path.splitext(fn)[-1].lower() in file_extension_handlers.keys()]
	if not files:
		raise click.ClickException("No data files found in folder {0}".format(click.format_filename(path)))
	
	with click.progressbar(files, label='Processing files:', item_show_func=get_filename, color='green') as bar:
		for filepath in bar:
			file_results = _process_file(filepath, plate, blank_strain, ref_strain, max_time, guess, param_min, param_max, param_fix, weights, ci, nsamples)
			results.extend(file_results)
	
	output_table = pd.DataFrame(results)
	output_table.to_csv(output_file, index=False)
	if VERBOSE and output_file.name != '-':
		click.secho("Wrote output to %s" % output_file.name, fg='green')


def _load_file(filepath, max_time, plate):
	fn, ext = os.path.splitext(filepath)
	echo_info("\tHandler: {1}\n".format(filepath, ext))
	handler = file_extension_handlers.get(ext)
	if handler is None:
		echo_info("No handler found for file {0}".format(click.format_filename(filepath)))
		return None
	try:
		if np.isfinite(max_time):
			df = handler(filepath, plate=plate, max_time=max_time)
		else:
			df = handler(filepath, plate=plate)
	except IOError as e:
		ioerror_to_click_exception(e)
	except (InvalidFileException, zipfile.BadZipFile, OSError, ValueError) as e:
		raise click.FileError(filepath, hint="parser error, probably not a {1} file, {0}".format(e, ext))
	return df


def _process_file(filepath, plate, blank_strain, ref_strain, max_time, guess, param_min, param_max, param_fix, weights, ci, nsamples):
	"""Analyses a single growth curves file.

	See also
	--------
	analyse
	"""
	results = []
	fn, ext = os.path.splitext(filepath)
	df = _load_file(filepath, max_time, plate)
	if df is None:
		return results

	strains = plate.Strain.unique().tolist()

	if blank_strain is not None and blank_strain != 'none': 
		if blank_strain in strains:
			bg = df[(df.Strain == blank_strain) & (df.Time == df.Time.min())]
			bg = bg.OD.mean()
			df.OD -= bg
			df.loc[df.OD < 0, 'OD'] = 0
		else:
			echo_error("Warning! Blank strain '%s' doesn't exist" % blank_strain)

	if PLOT:
		wells_plot_fn = fn + '_wells.png'
		g = curveball.plots.plot_wells(df, output_filename=wells_plot_fn)
		echo_info("Wrote wells plot to %s" % click.format_filename(wells_plot_fn))

		strains_plot_fn = fn + '_strains.png'
		g = curveball.plots.plot_strains(df, output_filename=strains_plot_fn)
		echo_info("Wrote strains plot to %s" % click.format_filename(strains_plot_fn))
	
	if blank_strain in strains: 
		strains.remove(blank_strain)
	if ref_strain in strains:
		strains.remove(ref_strain)
		strains.insert(0, ref_strain)
	else:
		echo_error("Warning, reference strains '%s' doesn't exist!" % ref_strain)

	for strain in strains:
		strain_df = df[df.Strain == strain]
		if strain_df.OD.nunique() <= 1:
			warnings.warn("Strain %s has constant OD values; skipping fit." % strain)
			res = {
				'folder': os.path.dirname(filepath),
				'filename': os.path.splitext(os.path.basename(fn))[0],
				'strain': strain,
				'model': None,
				'RSS': np.nan,
				'RMSD': np.nan,
				'NRMSD': np.nan,
				'CV(RMSD)': np.nan,
				'bic': np.nan,
				'aic': np.nan,
				'weighted_bic': np.nan,
				'weighted_aic': np.nan,
				'y0': np.nan,
				'K': np.nan,
				'r': np.nan,
				'nu': np.nan,
				'q0': np.nan,
				'v': np.nan,
				'has_lag': False,
				'has_nu': False,
				'max_growth_rate': np.nan,
				'min_doubling_time': np.nan,
				'lag': np.nan,
				'w': np.nan
			}
			results.append(res)
			plt.clf()
			continue
		_ = curveball.models.fit_model(strain_df, param_guess=guess, param_min=param_min, param_max=param_max, param_fix=param_fix, use_weights=weights, PLOT=PLOT, PRINT=VERBOSE)
		if PLOT:
			fit_results,fig,ax = _
			strain_plot_fn = fn + ('_strain_%s.png' % strain)
			fig.savefig(strain_plot_fn)
			echo_info("Wrote strain %s plot to %s" % (strain, click.format_filename(strain_plot_fn)))
		else:
			fit_results = _

		res = {}
		fit = fit_results[0]
		res['folder'] = os.path.dirname(filepath)
		res['filename'] = os.path.splitext(os.path.basename(fn))[0]
		res['strain'] = strain
		res['model'] = fit.model.name
		res['RSS'] = fit.chisqr
		res['RMSD'] = np.sqrt(res['RSS'] / fit.ndata)
		od_range = strain_df.OD.max() - strain_df.OD.min()
		if od_range == 0:
			warnings.warn("Strain %s has constant OD; normalized metrics undefined" % strain)
			res['NRMSD'] = np.nan
		else:
			res['NRMSD'] = res['RMSD'] / od_range
		od_mean = strain_df.OD.mean()
		if od_mean == 0:
			warnings.warn("Strain %s has mean OD zero; CV undefined" % strain)
			res['CV(RMSD)'] = np.nan
		else:
			res['CV(RMSD)'] = res['RMSD'] / od_mean
		res['bic'] = fit.bic
		res['aic'] = fit.aic
		res['weighted_bic'] = fit.weighted_bic
		res['weighted_aic'] = fit.weighted_aic
		params = fit.params
		res['y0'] = params['y0'].value
		res['K'] = params['K'].value
		res['r'] = params['r'].value
		res['nu'] = params['nu'].value if 'nu' in params else 1
		res['q0'] = params['q0'].value if 'q0' in params else 0
		res['v'] = params['v'].value if 'v' in params else 0
		res['has_lag'] = curveball.models.has_lag(fit_results)
		res['has_nu'] = curveball.models.has_nu(fit_results, PRINT=VERBOSE)
		res['max_growth_rate'] = curveball.models.find_max_growth(fit)[-1]
		res['min_doubling_time'] = curveball.models.find_min_doubling_time(fit)
		res['lag'] = curveball.models.find_lag(fit)
		if ci:
			param_samples = curveball.models.bootstrap_params(strain_df, fit, nsamples=nsamples)
			_, _, low, high = curveball.models.find_max_growth_ci(fit, param_samples)
			res['max_growth_rate_low'] = low
			res['max_growth_rate_high'] = high			
			low, high = curveball.models.find_lag_ci(fit, param_samples)
			res['lag_low'] = low
			res['lag_high'] = high			
			low, high = curveball.models.find_min_doubling_time_ci(fit, param_samples)
			res['min_doubling_time_low'] = low
			res['min_doubling_time_high'] = high
			low, high = curveball.models.find_K_ci(param_samples)
			res['K_low'] = low
			res['K_high'] = high

		if strain == ref_strain:
			ref_fit = fit
			res['w'] = 1
		elif ref_strain in strains:
			colors = plate[plate.Strain.isin([strain, ref_strain])].Color.unique()
			_ = curveball.competitions.compete(fit, ref_fit, hours=df.Time.max(), colors=colors, PLOT=PLOT)
			if PLOT:
				t,y,fig,ax = _
				competition_plot_fn = fn + ('_%s_vs_%s.png' % (strain, ref_strain))
				fig.savefig(competition_plot_fn)
				echo_info("Wrote competition %s vs %s plot to %s" % (strain, ref_strain, click.format_filename(strain_plot_fn)))
			else:
				t,y = _
			res['w'] = curveball.competitions.fitness_LTEE(y, assay_strain=0, ref_strain=1)
			# TODO CI for w
		results.append(res)
		plt.clf()
	return results


if __name__ == '__main__':
    cli()
def _resource_path(package, name):
	return os.fspath(resources.files(package).joinpath(name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from __future__ import division
from builtins import range
from unittest import TestCase, main
import sys
import os
import numbers

import numpy as np
import curveball
import matplotlib
import pandas as pd

import logging
logging.getLogger('matplotlib').setLevel(logging.INFO)


CI = os.environ.get('CI', 'false').lower() == 'true'


def logistic(t, y0, r, K):
    return K / (1.0 - (1.0 - K / y0) * np.exp(-r * t))

class ODETestCase(TestCase):
    def test_double_baranyi_roberts_ode(self):
        for func_name in dir(curveball.competitions):
            if func_name.startswith('double_baranyi_roberts_ode'):
                func = getattr(curveball.competitions, func_name)
                dy1dt, dy2dt = func((1, 1), 1, (1, 1), (1, 1), (1, 1), (1, 1), (1, 1))
                self.assertIsInstance(dy1dt, numbers.Real)
                self.assertIsInstance(dy2dt, numbers.Real)
                self.assertTrue(np.isfinite(dy1dt))
                self.assertTrue(np.isfinite(dy2dt))


    def test_baranyi_roberts(self):
        for func_name in dir(curveball.competitions):
            if func_name.startswith('baranyi_roberts_'):
                func = getattr(curveball.competitions, func_name)
                dy1dt, dy2dt = func((1, 1), 1, (1, 1), (1, 1), (1, 1), (1, 1), (1, 1), (1, 1))
                self.assertIsInstance(dy1dt, numbers.Real)
                self.assertIsInstance(dy2dt, numbers.Real)
                self.assertTrue(np.isfinite(dy1dt))
                self.assertTrue(np.isfinite(dy2dt))



class CompetitionTestCase(TestCase):

    def setUp(self):
        t = np.linspace(0, 24, 20)
        y0 = 0.1
        r1, r2 = 0.3, 0.4
        K = 1.0

        y1 = logistic(t, y0, r1, K)
        y2 = logistic(t, y0, r2, K)

        model = curveball.baranyi_roberts_model.Logistic()
        params1 = model.guess(data=y1, t=t)
        self.m1 = model.fit(data=y1, t=t, params=params1)
        params2 = model.guess(data=y2, t=t)
        self.m2 = model.fit(data=y2, t=t, params=params2)
        g  = (pd.DataFrame({'Time': t, 'OD': y1 + y2 + np.random.normal(0, 0.01)}) for _ in range(10))
        self.df_mixed = pd.concat(g)


    def tearDown(self):
        pass


    def test_compete(self):
        t, y = curveball.competitions.compete(self.m1, self.m2, PLOT=False)
        self.assertEqual(t.shape[0], y.shape[0])
        self.assertEqual(y.shape[1], 2)
        self.assertTrue(y[-1,0] > y[0,0])
        self.assertTrue(y[-1,1] > y[0,1])


    def test_compete_with_params(self):
        K1 = self.m1.best_values['K']
        r2 = self.m2.best_values['r']
        t, y = curveball.competitions.compete(self.m1, self.m2, params1={'K': 1.0}, params2={'r': 0.01}, PLOT=False)
        self.assertEqual(t.shape[0], y.shape[0])
        self.assertEqual(y.shape[1], 2)
        self.assertTrue(y[-1,0] > y[0,0])
        self.assertTrue(y[-1,1] > y[0,1])
        self.assertEqual(self.m1.best_values['K'], K1)
        self.assertEqual(self.m2.best_values['r'], r2)

    ### FIXME
    # def test_compete_plot(self):
    #     t, y, fig, ax = curveball.competitions.compete(self.m1, self.m2, colors=['r','b'], PLOT=True)
    #     self.assertEqual(t.shape[0], y.shape[0])
    #     self.assertEqual(y.shape[1], 2)
    #     self.assertTrue(y[-1,0] > y[0,0])
    #     self.assertTrue(y[-1,1] > y[0,1])
    #     self.assertIsInstance(fig, matplotlib.figure.Figure)
    #     self.assertIsInstance(ax, matplotlib.axes.Axes)
    #     if not CI:
    #         func_name = sys._getframe().f_code.co_name
    #         fig.savefig(func_name + ".png")


    def test_compete_resample(self):
        nsamples = 100
        t, y = curveball.competitions.compete(self.m1, self.m2, nsamples=nsamples, PLOT=False)
        self.assertEqual(t.shape[0], y.shape[0])
        self.assertEqual(y.shape[1], 2)
        self.assertEqual(y.shape[2], nsamples)
        self.assertTrue((y[-1,0,:] > y[0,0,:]).all())
        self.assertTrue((y[-1,1,:] > y[0,1,:]).all())


    # def test_compete_plot_resample(self):
    #     nsamples = 100
    #     self.m1.covar = np.ones((3,3)) * 0.01
    #     self.m2.covar = np.ones((3,3)) * 0.01

    #     ## FIXME
    #     t, y, fig, ax = curveball.competitions.compete(self.m1, self.m2, nsamples=nsamples, PLOT=True)
    #     self.assertEqual(t.shape[0], y.shape[0])
    #     self.assertEqual(y.shape[1], 2)
    #     ## FIXME
    #     #self.assertEqual(y.shape[2], nsamples)
    #     # self.assertTrue((y[-1,0,:] > y[0,0,:]).all())
    #     # self.assertTrue((y[-1,1,:] > y[0,1,:]).all())
    #     self.assertIsInstance(fig, matplotlib.figure.Figure)
    #     self.assertIsInstance(ax, matplotlib.axes.Axes)
    #     if not CI:
    #         func_name = sys._getframe().f_code.co_name
    #         fig.savefig(func_name + ".png")


    def test_fit_and_compete(self):
        t, y, a = curveball.competitions.fit_and_compete(self.m1, self.m2, self.df_mixed, PLOT=False)
        self.assertEqual(t.shape[0], y.shape[0])
        self.assertEqual(y.shape[1], 2)
        self.assertTrue(y[-1,0] > y[0,0])
        self.assertTrue(y[-1,1] > y[0,1])
        self.assertEqual(len(a), 2)
        self.assertIsInstance(a[0], numbers.Real)
        self.assertIsInstance(a[1], numbers.Real)


class FitnessTestCase(TestCase):

    def test_fitness_LTEE(self):
        t = np.linspace(0, 24)
        y1 = logistic(t, 0.1, 0.3, 1)
        y2 = logistic(t, 0.1, 0.4, 1)
        y = np.array((y1, y2)).T
        shape = y.shape
        w = curveball.competitions.fitness_LTEE(y)
        self.assertEqual(y.shape, shape)
        self.assertTrue(1 < w < 1.1)


    def test_fitness_LTEE_ci(self):
        nsamples = 100
        def rand():
            return np.random.normal(0, 0.01)
        t = np.linspace(0, 24)
        y = np.zeros((len(t), 2, nsamples))
        for i in range(nsamples):
            y[:,0,i] = logistic(t, 0.1, 0.3 + rand(), 1 + rand())
            y[:,1,i] = logistic(t, 0.1, 0.4 + rand(), 1 + rand())
        shape = y.shape
        w, low, high = curveball.competitions.fitness_LTEE(y, ci=0.95)
        self.assertEqual(y.shape, shape)
        self.assertTrue(1 < w < 1.1)
        self.assertTrue(low < w < high)

    # commented out due to travis-ci failure on py36
    #     self.assertTrue(np.isfinite(svals).all())
	#		AssertionError: False is not true
    # def test_selection_coefs_ts(self):
    #     t = np.linspace(0, 24)
    #     y1 = logistic(t, 0.1, 0.3, 1)
    #     y2 = logistic(t, 0.1, 0.4, 1)
    #     y = np.array((y1, y2)).T
    #     shape = y.shape
    #     svals = curveball.competitions.selection_coefs_ts(t, y, PLOT=False)
    #     self.assertEqual(y.shape, shape)
    #     self.assertIsInstance(svals, np.ndarray)
    #     self.assertIsInstance(svals[0], numbers.Real)
    #     self.assertTrue(np.isfinite(svals).all())


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of curveball.
# https://github.com/yoavram/curveball

# Licensed under the MIT license:
# http://www.opensource.org/licenses/MIT-license
# Copyright (c) 2015, Yoav Ram <yoav@yoavram.com>
from unittest import TestCase, main
import os
import types
import numpy as np
import matplotlib as mpl
import curveball
from curveball.likelihood import *

CI = os.environ.get('CI', 'false').lower() == 'true'
RANDOM_SEED = int(os.environ.get('RANDOM_SEED', 0))


class LoglikTestCase(TestCase):


	def setUp(self):
		self.params = dict(y0=0.1, K=1, r=1, nu=1, q0=np.inf, v=np.inf)
		self.df = curveball.models.randomize(reps=30, random_seed=RANDOM_SEED, as_df=True,
											 **self.params)
		self.t = self.df.Time.unique()
		self.y = self.df.groupby(by='Time').OD.mean()
		self.y_sig = self.df.groupby(by='Time').OD.std()				

	def test_loglik(self):
		ll = loglik(self.t, self.y, self.y_sig, 
					f=curveball.baranyi_roberts_model.baranyi_roberts_function,
					**self.params)
		self.assertIsInstance(ll, float)


	def test_ridge_regularization_lam0(self):
		penalty = ridge_regularization(0, **self.params)
		self.assertIsInstance(penalty, types.FunctionType)
		self.assertEqual(penalty(**self.params), 0)
		self.assertEqual(penalty(K=0.6), 0)


	def test_ridge_regularization_lam1(self):
		penalty = ridge_regularization(1, **self.params)
		self.assertIsInstance(penalty, types.FunctionType)
		self.assertEqual(penalty(**self.params), 0)
		self.assertTrue(penalty(K=0.6) > 0)


	def test_loglik_penalty(self):
		penalty = ridge_regularization(1, K=0.7)
		ll0 = loglik(self.t, self.y, self.y_sig, 
					f=curveball.baranyi_roberts_model.baranyi_roberts_function,
					**self.params)
		ll1 = loglik(self.t, self.y, self.y_sig, 
					f=curveball.baranyi_roberts_model.baranyi_roberts_function,
					penalty=penalty,
					**self.params)
		self.assertTrue(ll0 > ll1)


	def test_loglik_r_nu(self):
		L = loglik_r_nu(np.logspace(-2, 2, 10), np.logspace(-2, 2, 10), 
			self.df, 
			f=curveball.baranyi_roberts_model.baranyi_roberts_function, 
			**self.params)
		self.assertIsInstance(L, np.ndarray)
		self.assertTrue(L.ndim, 2)
		self.assertTrue(L.shape, (10, 10))
		self.assertIsInstance(L[0,0], float)


	def test_loglik_r_q0(self):
		L = loglik_r_q0(np.logspace(-2, 2, 10), np.logspace(-2, 2, 10), 
			self.df, 
			f=curveball.baranyi_roberts_model.baranyi_roberts_function, 
			**self.params)
		self.assertIsInstance(L, np.ndarray)
		self.assertTrue(L.ndim, 2)
		self.assertTrue(L.shape, (10, 10))
		self.assertIsInstance(L[0,0], float)


	def test_plot_loglik(self):
		L = loglik_r_nu(np.logspace(-2, 2, 10), np.logspace(-2, 2, 10), 
			self.df, 
			f=curveball.baranyi_roberts_model.baranyi_roberts_function, 
			**self.params)
		fig, ax = plot_loglik([L, L], np.logspace(-2, 2, 10), np.logspace(-2, 2, 10), 
					xlabel='r', ylabel='nu', columns=2, 
					fig_title='test', normalize=True,
                	ax_titles=['1', '2'], cmap='jet', colorbar=True, 
                	ax_width=4, ax_height=4)
		self.assertIsInstance(fig, mpl.figure.Figure)
		self.assertEqual(ax.size, 2)


	def test_plot_model_loglik(self):
		m = curveball.models.fit_model(self.df, PLOT=False, PRINT=False, 
									   models=curveball.baranyi_roberts_model.BaranyiRoberts)
		m = m[0]
		fig, ax = plot_model_loglik(m, self.df)
		self.assertIsInstance(fig, mpl.figure.Figure)